#
# statistics_wdgts_plotting.py <mike@peacecorps.org.cv>, <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.statistics_pages.grades import Grades
from mnemosyne.libmnemosyne.statistics_pages.schedule import Schedule
from mnemosyne.libmnemosyne.statistics_pages.easiness import Easiness
from mnemosyne.libmnemosyne.statistics_pages.cards_added import CardsAdded
from mnemosyne.libmnemosyne.statistics_pages.cards_learned import CardsLearned
from mnemosyne.libmnemosyne.statistics_pages.retention_score import RetentionScore
from mnemosyne.libmnemosyne.ui_components.statistics_widget import StatisticsWidget


class PlotStatisticsWdgt(QtWidgets.QWidget, StatisticsWidget):

    """A canvas to plot graphs according to the data and display contained in a
    statistics page.

    """

    def __init__(self, parent, page, **kwds):
        super().__init__(**kwds)
        self.parent = parent
        self.page = page

    def activate(self):
        StatisticsWidget.activate(self)

        # Late import to speed up app startup.
        import warnings
        warnings.filterwarnings("ignore", "(?s).*MATPLOTLIBDATA.*",
                                category=UserWarning)
        from matplotlib import use
        use("Qt5Agg")
        from matplotlib import rcParams
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg \
            as FigureCanvas

        self.setMinimumSize(640, 480)
        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding,
                           QtWidgets.QSizePolicy.MinimumExpanding)
        colour = self._background_colour(self.parent)
        fig = Figure(facecolor=colour, edgecolor=colour)
        self.canvas = FigureCanvas(fig)
        self.vbox_layout = QtWidgets.QVBoxLayout(self)
        self.vbox_layout.addWidget(self.canvas)
        self.canvas.setMinimumSize(640, 480)
        self.canvas.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding,
                                  QtWidgets.QSizePolicy.MinimumExpanding)
        self.canvas.setParent(self)
        self.axes = fig.add_subplot(111)
        self.canvas.updateGeometry()
        if self.config()["ui_language"].lower().startswith("zh"):
            rcParams['font.sans-serif'] = \
                ["Microsoft YaHei", "WenQuanYi Zen Hei", "Ume P Gothic O5"]

    def display_message(self, text):
        self.axes.clear()
        self.axes.set_xticklabels("")
        self.axes.set_yticklabels("")
        self.axes.set_xticks((0.0,))
        self.axes.set_yticks((0.0,))
        self.axes.text(0.5, 0.5, text, transform=self.axes.transAxes,
                       horizontalalignment="center", verticalalignment="center")

    def integers_only(self, x, pos=None):

        """Formatter to have only integer values on the axis."""

        if x == int(x):
            return "%d" % x
        else:
            return ""

    def _add_numbers_to_bars(self):
        # Add exact numeric value above each bar. The text padding is based
        # on the average height of non-zero bars. This gives a reasonable
        # padding for most data, but looks ugly when one bar is *much* larger
        # than the others. Note that bool() can be used as the filter function
        # since bool(0) == False.
        non_zero = list(filter(bool, self.page.y))
        if non_zero:
            avg = lambda s: sum(s) / len(s)
            avg_height = avg(list(filter(bool, self.page.y)))
        else:
            avg_height = 0
        pad = avg_height * 1.05 - avg_height
        for i in range(len(self.page.y)):
            height = self.page.y[i]
            if height == 0:
                continue
            self.axes.text(self.page.x[i], height + pad, "%d" % height,
                           ha="center", va="bottom", fontsize="small")

    def _background_colour(self, parent):

        """Return the parent's background colour.

        Sadly, this won't work on OS X, XP, or Vista since they use native
        theme engines for drawing, rather than the palette. See:
        http://doc.trolltech.com/4.4/qapplication.html#setPalette

        """

        if parent.style().objectName() == "macintosh (Aqua)":
            return "0.91"
        else:
            r, g, b, a = parent.palette().color(parent.backgroundRole()).getRgb()
            return [x / 255.0 for x in (r, g, b)]


class BarChartDaysWdgt(PlotStatisticsWdgt):

    colour = "blue"

    def show_statistics(self, variant):
        self.activate()
        # Determine variant-dependent formatting.
        if not self.page.y:
            self.display_message(_("No stats available."))
            return
        ticklabels_pos = lambda i, j, k: ["+%d" % x for x in range(i, j, k)]
        ticklabels_neg = lambda i, j, k: ["%d" % x for x in range(i, j, k)]
        if hasattr(self.page, "NEXT_WEEK") and \
               variant == self.page.NEXT_WEEK:
            xticks = list(range(1, 8, 1))
            xticklabels = ticklabels_pos(1, 8, 1)
            show_text_value = True
            linewidth = 1
        elif hasattr(self.page, "NEXT_MONTH") and \
            variant == self.page.NEXT_MONTH:
            xticks = [1] + list(range(5, 32, 5))
            xticklabels = ["+1"] + ticklabels_pos(5, 32, 5)
            show_text_value = False
            linewidth = 1
        elif hasattr(self.page, "NEXT_3_MONTHS") and \
            variant == self.page.NEXT_3_MONTHS:
            xticks = [1] + list(range(10, 92, 10))
            xticklabels = ["+1"] + ticklabels_pos(10, 92, 10)
            show_text_value = False
            linewidth = 1
        elif hasattr(self.page, "NEXT_6_MONTHS") and \
            variant == self.page.NEXT_6_MONTHS:
            xticks = [1] + list(range(30, 183, 30))
            xticklabels = ["+1"] + ticklabels_pos(30, 183, 30)
            show_text_value = False
            linewidth = 0
        elif hasattr(self.page, "NEXT_YEAR") and \
                 variant == self.page.NEXT_YEAR:
            xticks = [1] + list(range(60, 365, 60))
            xticklabels = ["+1"] + ticklabels_pos(60, 365, 60)
            show_text_value = False
            linewidth = 0
        elif hasattr(self.page, "LAST_WEEK") and \
            variant == self.page.LAST_WEEK:
            xticks = list(range(-7, 1, 1))
            xticklabels = ticklabels_neg(-7, 1, 1)
            show_text_value = True
            linewidth = 1
        elif hasattr(self.page, "LAST_MONTH") and \
            variant == self.page.LAST_MONTH:
            xticks = list(range(-30, -4, 5)) + [0]
            xticklabels = ticklabels_neg(-30, -4, 5) + ["0"]
            show_text_value = False
            linewidth = 1
        elif hasattr(self.page, "LAST_3_MONTHS") and \
            variant == self.page.LAST_3_MONTHS:
            xticks = list(range(-90, -9, 10)) + [0]
            xticklabels = ticklabels_neg(-90, -9, 10) + ["0"]
            show_text_value = False
            linewidth = 1
        elif hasattr(self.page, "LAST_6_MONTHS") and \
            variant == self.page.LAST_6_MONTHS:
            xticks = list(range(-180, -19, 20)) + [0]
            xticklabels = ticklabels_neg(-180, -19, 20) + ["0"]
            show_text_value = False
            linewidth = 0
        elif hasattr(self.page, "LAST_YEAR") and \
            variant == self.page.LAST_YEAR:
            xticks = list(range(-360, -59, 60)) + [0]
            xticklabels = ticklabels_neg(-360, -59, 60) + ["0"]
            show_text_value = False
            linewidth = 0
        else:
            raise AttributeError("Invalid variant")
        # Plot data.
        self.axes.bar(self.page.x, self.page.y, width=1, align="center",
                      linewidth=linewidth, color=self.colour, alpha=0.75,
                      edgecolor="black")
        self.axes.set_title(self.title)
        self.axes.set_xlabel(_("Days"))
        self.axes.set_xticks(xticks)
        self.axes.set_xticklabels(xticklabels)
        xmin, xmax = min(self.page.x), max(self.page.x)
        self.axes.set_xlim(xmin=xmin - 0.5, xmax=xmax + 0.5)
        # Quick and dirty fix to have retention plot always max out at 100%,
        # except for LAST_WEEK, when we need space to put the numbers.
        if variant == self.page.LAST_WEEK or not hasattr(self, "y_max"):
            self.axes.set_ylim(ymax=int(max(self.page.y) *  1.1) + 1)
        else:
            self.axes.set_ylim(ymax=self.y_max)
        from matplotlib.ticker import FuncFormatter
        self.axes.yaxis.set_major_formatter(FuncFormatter(self.integers_only))
        if show_text_value:
            self._add_numbers_to_bars()


class ScheduleWdgt(BarChartDaysWdgt):

    title = _("Number of cards scheduled")
    used_for = Schedule

    def activate(self):
        super().activate()
        self.title = _(self.title)


class RetentionScoreWdgt(BarChartDaysWdgt):

    used_for = RetentionScore
    colour = "green"
    title = _("Retention score (%)")
    y_max = 100

    def activate(self):
        super().activate()
        self.title = _(self.title)


class CardsAddedWdgt(BarChartDaysWdgt):

    used_for = CardsAdded
    colour = "red"
    title = _("Number of cards added")

    def activate(self):
        super().activate()
        self.title = _(self.title)


class CardsLearnedWdgt(BarChartDaysWdgt):

    used_for = CardsLearned
    colour = "yellow"
    title = _("Number of new cards learned")

    def activate(self):
        super().activate()
        self.title = _(self.title)


class GradesWdgt(PlotStatisticsWdgt):

    used_for = Grades

    def show_statistics(self, variant):
        self.activate()
        if not self.page.y:
            self.display_message(_("No stats available."))
            return
        self.axes.bar(self.page.x, self.page.y, width=0.7, align="center",
            color="blue", alpha=0.75)
        self.axes.set_title(_("Number of cards"))
        self.axes.set_xlabel(_("Grades"))
        self.axes.set_xticks(self.page.x)
        self.axes.set_xticklabels([_("Not seen")] + list(range(0, 6)))
        xmin, xmax = min(self.page.x), max(self.page.x)
        self.axes.set_xlim(xmin=xmin - 0.5, xmax=xmax + 0.5)
        self.axes.set_ylim(ymax=int(max(self.page.y) *  1.1) + 1)
        from matplotlib.ticker import FuncFormatter
        self.axes.yaxis.set_major_formatter(FuncFormatter(self.integers_only))
        self._add_numbers_to_bars()


class EasinessWdgt(PlotStatisticsWdgt):

    used_for = Easiness

    def show_statistics(self, variant):
        self.activate()
        if not self.page.data:
            self.display_message(_("No stats available."))
            return
        n, bins, patches = self.axes.hist(self.page.data, range=(1.3, 3.7),
            bins=24, align="left", facecolor="red", alpha=0.75)
        self.axes.set_title(_("Number of cards"))
        self.axes.set_xlabel(_("Easiness"))
        self.axes.set_xlim(xmin=1.3, xmax=3.7)
        self.axes.set_ylim(ymax=int(max(n) *  1.1) + 1)
        from matplotlib.ticker import FuncFormatter
        self.axes.yaxis.set_major_formatter(FuncFormatter(self.integers_only))



#
# translator_dlg.py <Peter.Bienstman@UGent.be>
#

import os
import shutil

from PyQt5 import QtGui, QtCore, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.utils import traceback_string
from mnemosyne.libmnemosyne.utils import expand_path, contract_path
from mnemosyne.libmnemosyne.ui_components.dialogs import TranslatorDialog
from mnemosyne.pyqt_ui.ui_translator_dlg import Ui_TranslatorDlg



class DownloadThread(QtCore.QThread):

    finished_signal = QtCore.pyqtSignal(str)
    error_signal = QtCore.pyqtSignal(str)

    def __init__(self, translator, card_type, foreign_text, target_language_id):
        super().__init__()
        self.translator = translator
        self.card_type = card_type
        self.foreign_text = foreign_text
        self.target_language_id = target_language_id

    def run(self):
        try:
            translation = self.translator.translate(\
                self.card_type, self.foreign_text, self.target_language_id)
            if translation:
                self.finished_signal.emit(translation)
            else:
                self.error_signal.emit(_("Could not contact Google servers..."))
        except Exception as e:
            self.error_signal.emit(str(e) + "\n" + traceback_string())


class TranslatorDlg(QtWidgets.QDialog, TranslatorDialog, Ui_TranslatorDlg):

    def __init__(self, translator, **kwds):
        self.translator = translator
        super().__init__(**kwds)
        self.setupUi(self)
        self.last_foreign_text = ""
        self.text_to_insert = ""

    def activate(self, card_type, foreign_text):
        TranslatorDialog.activate(self)
        # Set text and font for the foreign text.
        self.card_type = card_type
        fact_key = self.config().card_type_property(\
            "foreign_fact_key", card_type, default="")
        font_string = self.config().card_type_property(\
            "font", card_type, fact_key)
        if font_string:
            font = QtGui.QFont()
            font.fromString(font_string)
            self.foreign_text.setCurrentFont(font)
        self.foreign_text.setPlainText(foreign_text)
        # Set target language.
        self.target_language_id = self.config().card_type_property(\
            "translation_language_id", card_type, default="en")
        self.language_id_with_name = {}
        for language in self.languages():
            self.language_id_with_name[language.name] = language.used_for
            self.target_languages.addItem(language.name)
            if language.used_for == self.target_language_id:
                saved_index = self.target_languages.count()-1
        self.target_languages.setCurrentIndex(saved_index)
        # Only now it's safe to connect to the slot.
        self.target_languages.currentIndexChanged.connect(\
            self.target_language_changed)
        # Auto download.
        self.insert_button.setEnabled(False)
        self.download_translation()
        self.exec_()

    def foreign_text_changed(self):
        # Force the user to preview.
        self.insert_button.setEnabled(False)
        self.preview_button.setDefault(True)

    def target_language_changed(self):
        self.target_language_id = self.language_id_with_name[\
            self.target_languages.currentText()]
        self.config().set_card_type_property("translation_language_id",
            self.target_language_id, self.card_type)
        self.download_translation()

    def download_translation(self):
        self.last_foreign_text = self.foreign_text.toPlainText()
        if not self.last_foreign_text:
            return
        # Note that we need to save the QtThread as a class variable,
        # otherwise it will get garbage collected.
        self.download_thread = DownloadThread(self.translator, self.card_type,
            self.last_foreign_text, self.target_language_id)
        self.download_thread.finished_signal.connect(self.show_translation)
        self.download_thread.error_signal.connect(self.show_error)
        self.main_widget().set_progress_text(_("Downloading..."))
        self.download_thread.start()

    def show_error(self, error_message):
        self.main_widget().close_progress()
        self.main_widget().show_error(error_message)
        
    def show_translation(self, translation):
        self.main_widget().close_progress()
        self.translated_text.setPlainText(translation)
        self.insert_button.setEnabled(True)
        self.insert_button.setDefault(True)
        self.insert_button.setFocus(True)

    def preview(self):
        if self.foreign_text.toPlainText() != self.last_foreign_text:
            self.download_translation()

    def accept(self):
        self.text_to_insert = self.translated_text.toPlainText()
        QtWidgets.QDialog.accept(self)



#
# logger.py <Peter.Bienstman@UGent.be>
#

import os
import time

from mnemosyne.libmnemosyne.component import Component


class Logger(Component):

    component_type = "log"

    def __init__(self, component_manager):
        self.active = False
        Component.__init__(self, component_manager)

    def activate(self):
        Component.activate(self)
        self._timestamp = None
        self.upload_thread = None
        self.archive_old_log()
        self.start_logging()
        if self.config()["upload_science_logs"]:
            from mnemosyne.libmnemosyne.log_uploader import LogUploader
            self.upload_thread = LogUploader(self.component_manager)
            self.upload_thread.start()
        self.active = True

    def get_timestamp(self):

        """If self._timestamp == None (the default), then the timestamp will
        be the current time. It is useful to be able to override this, e.g.
        during database import or syncing, when you need to add log entries
        to the database that happened in the past.

        """

        if self._timestamp == None:
            return time.time()
        else:
            return self._timestamp

    def set_timestamp(self, timestamp):
        self._timestamp = timestamp

    timestamp = property(get_timestamp, set_timestamp)

    def start_logging(self):
        pass

    def started_program(self, version_string=None):
        pass

    def stopped_program(self):
        pass

    def started_scheduler(self, scheduler_name=None):
        pass

    def loaded_database(self, machine_id=None, scheduled_count=None,
        non_memorised_count=None, active_count=None):
        pass

    def saved_database(self, machine_id=None, scheduled_count=None,
        non_memorised_count=None, active_count=None):
        pass

    def added_card(self, card):
        pass

    def edited_card(self, card):
        pass

    def deleted_card(self, card):
        pass

    def repetition(self, card, scheduled_interval, actual_interval,
        thinking_time):
        pass

    def added_tag(self, tag):
        pass

    def edited_tag(self, tag):
        pass

    def deleted_tag(self, tag):
        pass

    def added_media_file(self, filename):
        pass

    def edited_media_file(self, filename):
        pass

    def deleted_media_file(self, filename):
        pass

    def added_fact(self, fact):
        pass

    def edited_fact(self, fact):
        pass

    def deleted_fact(self, fact):
        pass

    def added_fact_view(self, fact_view):
        pass

    def edited_fact_view(self, fact_view):
        pass

    def deleted_fact_view(self, fact_view):
        pass

    def added_card_type(self, card_type):
        pass

    def edited_card_type(self, card_type):
        pass

    def deleted_card_type(self, card_type):
        pass

    def added_criterion(self, criterion):
        pass

    def edited_criterion(self, criterion):
        pass

    def deleted_criterion(self, criterion):
        pass

    def edited_setting(self, key):
        pass

    def dump_to_science_log(self):
        pass

    def log_index_of_last_upload(self):

        """We don't store this info in the configuration, but determine it on
        the fly, so that users can copy configuration files between their
        machines.

        1.x log names have the format userid_index.bz2.
        2.x log names have the format userid_machineid_index.bz2

        Obviously, we should only consider the logs from our own machine.

        """

        _dir = os.listdir(os.path.join(self.config().data_dir, "history"))
        history_files = [x for x in _dir if x[-4:] == ".bz2"]
        max_log_index = 0
        this_machine_id = self.config().machine_id()
        for history_file in history_files:
            user_and_machine, log_index_and_suffix = history_file.rsplit("_", 1)
            if "_" in user_and_machine:
                user, machine = user_and_machine.split("_")
                if machine != this_machine_id:
                    continue
            log_index = int(log_index_and_suffix.split(".")[0])
            if log_index > max_log_index:
                max_log_index = log_index
        return max_log_index

    def archive_old_log(self):

        """Archive log to history folder if it's large enough."""

        if not self.config()["upload_science_logs"]:
            return
        data_dir = self.config().data_dir
        log_name = os.path.join(data_dir, "log.txt")
        try:
            log_size = os.stat(log_name).st_size
        except:
            log_size = 0
        if log_size > self.config()["max_log_size_before_upload"]:
            user = self.config()["user_id"]
            machine = self.config().machine_id()
            index = self.log_index_of_last_upload() + 1
            archive_name = "%s_%s_%05d.bz2" % (user, machine, index)
            import bz2  # Not all platforms have bz.
            f = bz2.open(os.path.join(data_dir, "history",
                archive_name), "w")
            for l in open(log_name):
                f.write(l.encode("utf-8"))
            f.close()
            os.remove(log_name)

    def deactivate(self):
        if self.upload_thread:
            from mnemosyne.libmnemosyne.gui_translator import _
            print((_("Waiting for uploader thread to stop...").encode("utf-8")))
            self.upload_thread.join()
            print((_("Done!").encode("utf-8")))

    def warn_too_many_cards(self):
        pass



#
# database.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component


class Database(Component):

    """Interface class describing the functions to be implemented by the
    actual database classes.

    Apart from the basic interface defined here, depending on the situation
    a database can also implement functions for logging, statistics and
    syncing (see SQLite_logging.py, SQLite_statistics.py, SQLite_sync.py).

    """

    version = ""
    default_name = "default"  # Without suffix, should not be translated.
    default_criterion_name = "__DEFAULT__"
    suffix = ""
    component_type = "database"

    def deactivate(self):
        Component.deactivate(self)
        self.unload()

    def path(self):

        """Returns full path of the database."""

        raise NotImplementedError

    def data_dir(self):

        """Returns directory of the database."""

        raise NotImplementedError

    def name(self):

        """Returns name of the database, without parent paths, but with
        extensions.

        """

        raise NotImplementedError

    def display_name(self):

        """Returns bare name of the database, without parent paths and
        without extension.

        """

        raise NotImplementedError

    # File operations.

    def release_connection(self):

        """Release the connection, so that it may be recreated in a separate
        thread.

        """

        raise NotImplementedError

    def new(self, path):
        raise NotImplementedError

    def save(self, path=None):
        raise NotImplementedError

    def backup(self):
        raise NotImplementedError

    def restore(self, path):
        raise NotImplementedError

    def load(self, path):
        raise NotImplementedError

    def unload(self):
        raise NotImplementedError

    def abandon(self):
        raise NotImplementedError

    def is_loaded(self):
        raise NotImplementedError

    def is_empty(self):
        raise NotImplementedError

    # Functions to conform to openSM2sync API.

    def user_id(self):
        return self.config()["user_id"]

    def change_user_id(self, user_id):
        self.config().change_user_id(user_id)

    # Tags.

    def add_tag(self, tag):
        raise NotImplementedError

    def tag(self, id, is_id_internal):
        raise NotImplementedError

    def update_tag(self, tag):
        raise NotImplementedError

    def delete_tag(self, tag):
        raise NotImplementedError

    def get_or_create_tag_with_name(self, name):
        raise NotImplementedError

    def get_or_create_tags_with_names(self, names):
        raise NotImplementedError

    def delete_tag_if_unused(self, tag):
        raise NotImplementedError

    def tags(self):
        raise NotImplementedError

    def has_tag_with_id(self, id):
        return NotImplementedError

    # Facts.

    def add_fact(self, fact):
        raise NotImplementedError

    def fact(self, id, is_id_internal):
        raise NotImplementedError

    def update_fact(self, fact):
        raise NotImplementedError

    def delete_fact(self, fact):
        raise NotImplementedError

    def has_fact_with_id(self, id):
        return NotImplementedError

    # Cards.

    def add_card(self, card):
        raise NotImplementedError

    def card(self, id, is_id_internal):
        raise NotImplementedError

    def update_card(self, card, repetition_only=False):
        raise NotImplementedError

    def delete_card(self, card):
        raise NotImplementedError

    def tags_from_cards_with_internal_ids(self, _card_ids):
        raise NotImplementedError

    def add_tag_to_cards_with_internal_ids(self, tag, _card_ids):
        raise NotImplementedError

    def remove_tag_from_cards_with_internal_ids(self, tag, _card_ids):
        raise NotImplementedError

    def has_card_with_id(self, id):
        return NotImplementedError

    # Fact views.

    def add_fact_view(self, fact_view):
        raise NotImplementedError

    def fact_view(self, id, is_id_internal):
        raise NotImplementedError

    def update_fact_view(self, fact_view):
        raise NotImplementedError

    def delete_fact_view(self, fact_view):
        raise NotImplementedError

    def has_fact_view_with_id(self, id):
        return NotImplementedError

    # Card types.

    def add_card_type(self, card_type):
        raise NotImplementedError

    def card_type(self, id, is_id_internal):
        raise NotImplementedError

    def is_user_card_type(self, card_type):
        raise NotImplementedError

    def is_in_use(self, card_type):
        raise NotImplementedError

    def has_clone(self, card_type):
        raise NotImplementedError

    def update_card_type(self, card_type):
        raise NotImplementedError

    def delete_card_type(self, card_type):
        raise NotImplementedError

    def has_card_type_with_id(self, id):
        return NotImplementedError

    # Criteria.

    def add_criterion(self, criterion):
        raise NotImplementedError

    def criterion(self, id, is_id_internal):
        raise NotImplementedError

    def update_criterion(self, criterion):
        raise NotImplementedError

    def delete_criterion(self, criterion):
        raise NotImplementedError

    def set_current_criterion(self, criterion):
        raise NotImplementedError

    def current_criterion(self):
        raise NotImplementedError

    def criteria(self):
        raise NotImplementedError

    def has_criterion_with_id(self, id):
        return NotImplementedError

    # Queries.

    def cards_from_fact(self, fact):

        """Return a list of the cards deriving from a fact."""

        raise NotImplementedError

    def duplicates_for_fact(self, fact, card_type):

        """Return facts with same 'card_type.unique_fact_keys' data as 'fact'."""

        raise NotImplementedError

    def card_types_in_use(self):
        raise NotImplementedError

    # Card queries used by the scheduler. Returns tuples of internal ids
    # (_card_id, _fact_id) Should function as an iterator in order to save
    # memory. "sort_key" is a string of an attribute of Card to be used for
    # sorting, with "" standing for the order in which the cards where added
    # (no sorting), and "random" is used to shuffle the cards. "limit" is
    # used to limit the number of cards returned by the iterator, with -1
    # meaning no limit.

    def cards(self, sort_key="", limit=-1):
        raise NotImplementedError

    def cards_due_for_ret_rep(self, now, sort_key="", limit=-1):
        raise NotImplementedError

    def cards_to_relearn(self, grade, sort_key="", limit=-1):
        raise NotImplementedError

    def cards_new_memorising(self, grade, sort_key="", limit=-1):
        raise NotImplementedError

    def cards_unseen(self, sort_key="", limit=-1):
        raise NotImplementedError

    def cards_learn_ahead(self, now, sort_key="", limit=-1):
        raise NotImplementedError

    def recently_memorised_count(self, max_ret_reps):
        raise NotImplementedError

    # Extra queries for custom schedulers.

    def set_scheduler_data(self, scheduler_data):
        raise NotImplementedError

    def cards_with_scheduler_data(self, scheduler_data, sort_key="", limit=-1,
                                  max_ret_reps=-1):
        raise NotImplementedError

    def scheduler_data_count(self, scheduler_data, max_ret_reps=-1):
        raise NotImplementedError

    #
    # Extra queries for language analysis.
    #

    def known_recognition_questions_count_from_card_types_ids(\
        self, card_type_ids):
        raise NotImplementedError

    def known_recognition_questions_from_card_types_ids(self, card_type_ids):
        raise NotImplementedError

    def sorted_card_types(self):

        """Sorts card types so that all the built-in card types appear first,
        in the order determined by their id, and then all the user card types
        appear alphabetically.

        """

        result = []
        user_card_types = []

        for card_type in self.card_types():
            if self.is_user_card_type(card_type):
                user_card_types.append(card_type);
            else:
                result.append(card_type);

        result.sort(key=lambda x: x.id)
        user_card_types.sort(key=lambda x: x.name.lower())

        result.extend(user_card_types)

        return result



class DatabaseMaintenance(Component):

    """This component performs automatic database maintenance (like
    archiving of old logs) and can be run from the UI or automatically from
    the controller.

    This version is unthreaded, and is OK for running on a headless server
    (which has no UI to interrupt) and for Android (since the entire backend
    runs in thread there anyhow).

    """

    component_type = "database_maintenance"

    def run(self):
        self.main_widget().set_progress_text(_("Compacting database..."))
        self.database().archive_old_logs()
        self.database().defragment()
        self.main_widget().close_progress()




#
# SQLite_statistics.py <Peter.Bienstman@UGent.be>
#

import time
import datetime

from openSM2sync.log_entry import EventTypes

HOUR = 60 * 60 # Seconds in an hour.
DAY = 24 * HOUR # Seconds in a day.


class SQLiteStatistics(object):

    """Code to be injected into the SQLite database class through inheritance,
    so that SQLite.py does not becomes too large.

    """

    def tag_count(self):
        return self.con.execute("select count() from tags").fetchone()[0]

    def fact_count(self):
        return self.con.execute("select count() from facts").fetchone()[0]

    def card_count(self):
        return self.con.execute("""select count() from cards""").fetchone()[0]

    def non_memorised_count(self):
        return self.con.execute("""select count() from cards
            where active=1 and grade<2""").fetchone()[0]

    def scheduled_count(self, timestamp):
        count = self.con.execute("""select count() from cards
            where active=1 and grade>=2 and ?>=next_rep
            and ret_reps_since_lapse<=?""", (timestamp,
            self.config()["max_ret_reps_since_lapse"])).fetchone()[0]
        return count

    def active_count(self):
        return self.con.execute("""select count() from cards
            where active=1""").fetchone()[0]

    def easinesses(self, active_only):
        query = "select easiness from cards where grade>=0"
        if active_only:
            query += " and active=1"
        return [cursor[0] for cursor in self.con.execute(query)]

    def easinesses_for_tag(self, tag, active_only):
        query = """select cards.easiness from cards, tags_for_card where
            tags_for_card._card_id=cards._id and cards.grade>=0 and
            tags_for_card._tag_id=?"""
        if active_only:
            query += " and cards.active=1"
        return [cursor[0] for cursor in self.con.execute(query,
            (tag._id, ))]

    def card_count_for_fact_view(self, fact_view, active_only):
        query = "select count() from cards where fact_view_id=?"
        if active_only:
            query += " and active=1"
        return self.con.execute(query, (fact_view.id, )).fetchone()[0]

    def card_count_for_grade(self, grade, active_only):
        query = "select count() from cards where grade=?"
        if active_only:
            query += " and active=1"
        return self.con.execute(query, (grade, )).fetchone()[0]

    def card_count_for_tags(self, tags, active_only):

        """ Determine the number of cards in a set of tags. Note that a card
        could have one or more of these tags.

        """

        if len(tags) == 0:
            return 0
        if active_only == True:
            raise NotImplementedError
        query = "select count(distinct _card_id) from tags_for_card where "
        args = []
        # Since tag typically contains 1 element, a query with 'or' is faster
        # than one with 'in ()'.
        for tag in tags:
            query += "_tag_id=? or "
            args.append(tag._id)
        query = query.rsplit("or ", 1)[0]
        return self.con.execute(query, args).fetchone()[0]

    def card_count_for_grade_and_tag(self, grade, tag, active_only):
        query = """select count() from cards, tags_for_card where
            tags_for_card._card_id=cards._id and tags_for_card._tag_id=?
            and grade=?"""
        if active_only:
            query += " and cards.active=1"
        return self.con.execute(query, (tag._id, grade)).fetchone()[0]

    def sister_card_count_scheduled_between(self, card, start, stop):

        """Return how many sister cards with grade >= 2 are scheduled at
        between 'start' (included) and 'stop' (excluded).

        """

        # TODO: profile to determine the most efficient version

        #return self.con.execute("""select count() from cards where active=1
        #    and grade>=2 and ?<=next_rep and next_rep<? and _id<>? and _id in
        #    (select _id from cards where _fact_id=?)""",
        #    (start, stop, card._id, card.fact._id)).fetchone()[0]

        #return self.con.execute("""select count() from cards where _id in
        #    (select _id from cards where _fact_id=?) and active=1
        #    and grade>=2 and ?<=next_rep and next_rep<? and _id<>?""",
        #    (card.fact._id, start, stop, card._id)).fetchone()[0]

        #_card_ids = [cursor[0] for cursor in self.con.execute(\
        #    "select _id from cards where _fact_id=?", (card.fact._id, ))]
        #query = "select count() from cards where _id in ("
        #for _card_id in _card_ids:
        #    query += str(_card_id) + ","
        #query = query[:-1] + """) and active=1 and grade>=2 and
        #    ?<=next_rep and next_rep<? and _id<>?"""
        #return self.con.execute(query, (start, stop, card._id)).fetchone()[0]

        _card_ids = [cursor[0] for cursor in self.con.execute(\
            "select _id from cards where _fact_id=?", (card.fact._id, ))]
        if len(_card_ids) == 1: # No sister cards
            return 0
        query = "select active, grade, next_rep from cards where _id in ("
        for _card_id in _card_ids:
            if _card_id != card._id:
                query += str(_card_id) + ","
        query = query[:-1] + """)"""
        count = 0
        for cursor in self.con.execute(query):
            if cursor[0] == True and cursor[1] >= 2 \
                and start <= cursor[2] < stop:
                count += 1
        return count

    def card_count_scheduled_between(self, start, stop):
        return self.con.execute(\
            """select count() from cards where grade>=2 and ?<=next_rep and
            next_rep<? and ret_reps_since_lapse<=? and active='1'""",
            (start, stop,
             self.config()["max_ret_reps_since_lapse"])).fetchone()[0]

    def start_of_day_n_days_ago(self, n):
        timestamp = time.time() - n * DAY \
                    - self.config()["day_starts_at"] * HOUR
        # Roll this back to the midnight before.
        date_only = datetime.date.fromtimestamp(timestamp) # Local date.
        start_of_day = int(time.mktime(date_only.timetuple()))
        # Bring back to 0:h with h="day_starts_at"
        start_of_day += self.config()["day_starts_at"] * HOUR
        return start_of_day

    def card_count_scheduled_n_days_ago(self, n):
        start_of_day = self.start_of_day_n_days_ago(n)
        actual_counts_for_machine = {}
        projected_counts_for_machine = {}
        # For each machine id, get the number of cards that were scheduled
        # that day. Make a distinction between the actual schedule and the
        # scheduled that was projected in the future during database load
        # events. For each machine, we take the largest number in the logs,
        # i.e. those at the start of the day.
        for cursor in self.con.execute("""select acq_reps, object_id from log
            where ?<=timestamp and timestamp<? and (event_type=? or
            event_type=?)""", (start_of_day, start_of_day + DAY,
            EventTypes.LOADED_DATABASE, EventTypes.SAVED_DATABASE)):
            count = cursor[0]
            machine = cursor[1]
            # Future projected schedule. Check if machine exists to deal with
            # Mnemosyne versions before 201203.
            if machine and machine.endswith(".fut"):
                if not machine in projected_counts_for_machine or \
                    (projected_counts_for_machine[machine] < count):
                    projected_counts_for_machine[machine] = count
            # Actual schedule.
            else:
                if not machine in actual_counts_for_machine or \
                    (actual_counts_for_machine[machine] < count):
                    actual_counts_for_machine[machine] = count
        # In case several machines report a different scheduded count, take
        # the minimum, as we assume that the larger number corresponds to
        # another machine which was kept running and therefore accumulated a
        # backlog, while the actual reviews happened on another machine.
        if len(actual_counts_for_machine) != 0:
            return min(actual_counts_for_machine.values())
        # In case there is no actual data, use the projected data, taking the
        # maximum over all possible machines.
        elif len(projected_counts_for_machine) != 0:
            return max(projected_counts_for_machine.values())
        # If there is no data, return 0 for unknown.
        else:
            return 0
        # Note that the algorithm above is still an approximation, e.g. there
        # is no way it can know about different cards sets that are active
        # during the day.
        # It can also go wrong after resolution of a sync conflict when all the
        # reviews are done for the day.

    def card_count_added_n_days_ago(self, n):
        start_of_day = self.start_of_day_n_days_ago(n)
        return self.con.execute(\
            """select count() from log where ?<=timestamp and timestamp<?
            and event_type=?""",
            (start_of_day, start_of_day + DAY, EventTypes.ADDED_CARD)).\
            fetchone()[0]

    def card_count_learned_n_days_ago(self, n):
        start_of_day = self.start_of_day_n_days_ago(n)
        return self.con.execute(\
            """select count() from log where ?<=timestamp and timestamp<?
            and event_type=? and grade>=2 and ret_reps==0""",
            (start_of_day, start_of_day + DAY, EventTypes.REPETITION)).\
            fetchone()[0]

    def retention_score_n_days_ago(self, n):
        start_of_day = self.start_of_day_n_days_ago(n)
        scheduled_cards_seen = self.con.execute(\
            """select count() from log where ?<=timestamp and timestamp<?
            and event_type=? and scheduled_interval!=0""",
            (start_of_day, start_of_day + DAY, EventTypes.REPETITION)).\
            fetchone()[0]
        if scheduled_cards_seen == 0:
            return 0
        scheduled_cards_correct = self.con.execute(\
            """select count() from log where ?<=timestamp and timestamp<?
            and event_type=? and scheduled_interval!=0 and grade>=2""",
            (start_of_day, start_of_day + DAY, EventTypes.REPETITION)).\
            fetchone()[0]
        return 100.0 * scheduled_cards_correct / scheduled_cards_seen

    def average_thinking_time(self, card):
        result = self.con.execute(\
            """select avg(thinking_time) from log where object_id=?
            and event_type=?""",
            (card.id, EventTypes.REPETITION)).fetchone()[0]
        if result:
            return result
        else:
            return 0

    def total_thinking_time(self, card):
        result = self.con.execute(\
            """select sum(thinking_time) from log where object_id=?
            and event_type=?""",
            (card.id, EventTypes.REPETITION)).fetchone()[0]
        if result:
            return result
        else:
            return 0



#
# current_card.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.statistics_page import HtmlStatisticsPage

DAY = 24 * 60 * 60 # Seconds in a day.

class CurrentCard(HtmlStatisticsPage):

    name = _("Current card")

    def prepare_statistics(self, variant):
        card = self.review_controller().card
        self.html = """<html<body>
        <style type="text/css">
        table { height: 100%;
                margin-left: auto; margin-right: auto;
                text-align: center}
        body  { background-color: white;
                margin: 0;
                padding: 0;
                border: thin solid #8F8F8F; }
        </style></head><table><tr><td>"""
        if not card:
            self.html += _("No current card.")
        elif card.grade == -1:
            self.html += _("Unseen card, no statistics available yet.")
        else:
            self.html += _("Grade") + ": %d<br>" % card.grade
            self.html += _("Easiness") + ": %1.2f<br>" % card.easiness
            self.html += _("Learning repetitions") + ": %d<br>" \
                % card.acq_reps
            self.html += _("Review repetitions") + ": %d<br>" \
                % card.ret_reps
            self.html += _("Lapses") + ": %d<br>" % card.lapses
            self.html += _("Last repetition") + ": %s<br>" \
                % self.scheduler().last_rep_to_interval_string(card.last_rep)
            self.html += _("Next repetition") + ": %s<br>" \
                % self.scheduler().next_rep_to_interval_string(card.next_rep)
            self.html += _("Average thinking time (secs)") + ": %d<br>" \
                % self.database().average_thinking_time(card)
            self.html += _("Total thinking time (secs)") + ": %d<br>" \
                % self.database().total_thinking_time(card)
        self.html += "</td></tr></table></body></html>"



#
# render_chain.py <Peter.Bienstman@UGent.be>
#

import copy
import string

from mnemosyne.libmnemosyne.component import Component


class RenderChain(Component):

    """A RenderChain details the operations needed to get from the raw data
    in a card to a representation of its question and answer, in a form either
    suitable for displaying in a browser, or exporting to a text file, ... .

    First the raw data is sent through Filters, which perform operations which
    can be useful for many card types, like expanding relative paths.

    Sometimes Mnemosyne will try to assemble the answer together with the
    question on a single page, e.g. if config()["QA_split"] == "single_window".
    For some Renderers where this does not make sense, e.g. the card browser,
    set 'never_join_q_to_a = True'.

    Then this data is assembled in the right order in a Renderer, which can be
    card type specific.

    'filters': list of Filter classes
    'renderers': list or Renderer classes

    Plugins can add Filters or Renderers for a new card type to a chain at run
    time. In case the rendering of a card type does not fit comfortably in the
    scheme below, an alternative is to override 'render_question' and
    'render_answer' in CardType directly.

    Each client should have a render chain with id="default" on startup.

    """

    component_type = "render_chain"
    id = "default"

    never_join_q_to_a = False

    filters = []
    renderers = []

    def __init__(self, component_manager):
        # To have an nice syntax when defining renderers, we do the
        # instantiation here.
        Component.__init__(self, component_manager)
        self._filters = []
        for filter in self.filters:
            self._filters.append(filter(component_manager))
        self._renderers = []
        self._renderer_for_card_type = {}
        for renderer in self.renderers:
            renderer = renderer(component_manager)
            self._renderers.append(renderer)
            self._renderer_for_card_type[renderer.used_for] = renderer

    def register_filter_at_front(self, filter_class, after=[]):

        """Register a filter at the very front of the render chain, but after
           a list of other filters already in the chain. The list should
           contain class names. (Using strings instead of classes means a
           plugin writer does not need to import the filters he wants to use
           in this list.)

           'filter_class' should be a class, not an instance.

        """

        filter = filter_class(self.component_manager)
        pos = 0
        for i in range(0, len(self._filters)):
            if self._filters[i].__class__.__name__ in after:
                pos = i + 1
        self._filters.insert(pos, filter)

    def register_filter_at_back(self, filter_class, before=[]):

        """Register a filter at the back of the render chain, but before
           a list of other filters already in the chain. The list should
           contain class names. (Using strings instead of classes means a
           plugin writer does not need to import the filters he wants to use
           in this list.)

           'filter_class' should be a class, not an instance.

        """

        filter = filter_class(self.component_manager)
        pos = len(self._filters)
        for i in range(len(self._filters) - 1, -1, -1):
            if self._filters[i].__class__.__name__ in before:
                pos = i
        self._filters.insert(pos, filter)

    def register_filter(self, filter_class, in_front=False):

        """'filter_class' should be a class, not an instance."""

        if in_front:
            self.register_filter_at_front(filter_class)
        else:
            self.register_filter_at_back(filter_class)

    def filter(self, filter_class):
        for filter_i in self._filters:
            if isinstance(filter_i, filter_class):
                return filter_i

    def unregister_filter(self, filter_class):

        """'filter_class' should be a class, not an instance."""

        for filter in self._filters:
            if isinstance(filter, filter_class):
                self._filters.remove(filter)
                break

    def register_renderer(self, renderer_class):

        """'renderer_class' should be a class, not an instance."""

        renderer = renderer_class(self.component_manager)
        self._renderer_for_card_type[renderer.used_for] = renderer

    def unregister_renderer(self, renderer_class):

        """'renderer_class' should be a class, not an instance."""

        for card_type, renderer in self._renderer_for_card_type.items():
            if isinstance(renderer, renderer_class):
                del self._renderer_for_card_type[card_type]
                break

    def renderer_for_card_type(self, card_type):
        if card_type in self._renderer_for_card_type:
            return self._renderer_for_card_type[card_type]
        if "::" in card_type.id:
            parent_id, child_id = card_type.id.rsplit("::", 1)
            parent = self.database().card_type(parent_id, is_id_internal=-1)
            return self.renderer_for_card_type(parent)
        return self._renderer_for_card_type[None]

    def render_question(self, card, **render_args):
        fact_keys = card.fact_view.q_fact_keys
        decorators = card.fact_view.q_fact_key_decorators
        if self.config()["QA_split"] == "single_window":
            render_args["align_top"] = True
        return self._render(card, fact_keys, decorators, **render_args)

    def render_answer(self, card, **render_args):
        fact_keys, decorators = [], {}
        a_on_top_of_q = render_args.get("a_on_top_of_q", False)
        if self.config()["QA_split"] == "single_window" and \
           not self.never_join_q_to_a and not a_on_top_of_q:
            render_args["align_top"] = True
            fact_keys += card.fact_view.q_fact_keys
            fact_keys.append("__line__")
            decorators.update(card.fact_view.q_fact_key_decorators)
        fact_keys += card.fact_view.a_fact_keys
        decorators.update(card.fact_view.a_fact_key_decorators)
        return self._render(card, fact_keys, decorators, **render_args)

    def _render(self, card, fact_keys, decorators, **render_args):
        # Note that the filters run only on the data, not on the full content
        # generated by the renderer, which would be much slower.
        fact_data = copy.copy(card.card_type.fact_data(card))
        for fact_key in fact_keys:
            if fact_key == "__line__":
                fact_data[fact_key] = "<hr id=answer>"
            if fact_key not in fact_data:  # Optional key.
                continue
            for filter in self._filters:
                fact_data[fact_key] = filter.run(fact_data[fact_key],
                    card, fact_key, **render_args)
            if fact_key in decorators:
                fact_data[fact_key] = string.Template(\
                    decorators[fact_key]).safe_substitute(fact_data)
        renderer = self.renderer_for_card_type(card.card_type)
        return renderer.render(\
            fact_data, fact_keys, card.card_type, **render_args)




#
# filter.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.component import Component


class Filter(Component):

    """Code which operates on a string and filters it to achieve extra
    functionality, e.g. converting relative paths to absolute paths.

    It is contained in a RenderChain and represents functionality which is
    useful for many different card types.

    The filters are executed in the order they are listed in the RenderChain.
    If you really need to make sure that your filter runs before the
    rest, set 'in_front=True' as argument in 'render_chain.register_filter'.
    
    """

    component_type = "filter"

    def run(self, text, card, fact_key, **render_args):
        raise NotImplementedError



#
# test_mnemosyne1xml_import.py <Johannes.Baiter@gmail.com>
#

import os
import sys
import shutil

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne import Mnemosyne
from openSM2sync.log_entry import EventTypes
from mnemosyne.libmnemosyne.ui_components.main_widget import MainWidget

last_error = ""
answer = None

class Widget(MainWidget):

    def show_information(self, message):
        if message.startswith("Warning: "):
            return 0
        if message.startswith("Ignoring unparsable file"):
            return 0
        if message.startswith("You appear"):
            return 0
        if message.startswith("Your queue is running empty, "):
            return 0
        raise NotImplementedError

    def show_error(self, message):
        global last_error
        last_error = message
        if message.startswith("These cards seem to have been imported before"):
            return
        if message.startswith("Unable to open"):
            return
        if message.startswith("Unable to parse"):
            return
        if message.startswith("Bad file version:"):
            return
        if message.startswith("XML file does not seem"):
            return
        raise NotImplementedError

    def show_question(self, question, option0, option1, option2):
        return answer


class TestMnemosyne1XMLImport(MnemosyneTest):

    def setup(self):
        self.initialise_data_dir()
        path = os.path.join(os.getcwd(), "..", "mnemosyne", "libmnemosyne",
                            "renderers")
        if path not in sys.path:
            sys.path.append(path)
        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
            asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.components.append(\
            ("test_mnemosyne1xml_import", "Widget"))
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.review_controller().reset()

    def xml_importer(self):
        for format in self.mnemosyne.component_manager.all("file_format"):
            if format.__class__.__name__ == "Mnemosyne1XML":
                return format

    def test_file_not_found(self):
        filename = os.path.join(os.getcwd(), "tests", "files", "nothere.xml")
        self.xml_importer().do_import(filename)
        assert last_error.startswith("Unable to open")

    def test_wrong_format(self):
        filename = os.path.join(os.getcwd(), "tests", "files", "wrong_format.xml")
        self.xml_importer().do_import(filename)
        assert last_error.startswith("Unable to parse")

    def test_bad_version(self):
        filename = os.path.join(os.getcwd(), "tests", "files", "bad_version.xml")
        self.xml_importer().do_import(filename)
        assert last_error.startswith("XML file does not seem")

    @MnemosyneTest.set_timezone_utc
    def test_card_type_1(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "1sided.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 4
        card = self.review_controller().card
        assert card.grade == 2
        assert card.easiness == 2.5
        assert card.acq_reps == 1
        assert card.ret_reps == 0
        assert card.lapses == 0
        assert card.acq_reps_since_lapse == 1
        assert card.ret_reps_since_lapse == 0
        assert [tag.name for tag in card.tags] == ["__UNTAGGED__"]
        assert card.last_rep == 1247529600
        assert card.next_rep == 1247616000
        assert card.id == "9cff728f"

    def test_card_type_1_abort(self):
        global answer
        answer = 1
        filename = os.path.join(os.getcwd(), "tests", "files", "1sided.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 0

    def test_card_type_1_unseen(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "1sided_unseen.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 1
        card = self.review_controller().card
        assert card.grade == -1
        assert card.easiness == 2.5
        assert card.acq_reps == 0
        assert card.ret_reps == 0
        assert card.lapses == 0
        assert card.acq_reps_since_lapse == 0
        assert card.ret_reps_since_lapse == 0
        assert card.last_rep == -1
        assert card.next_rep == -1
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_CARD, )).fetchone()[0] == 1

    def test_card_type_1_edited(self):
        filename = os.path.join(os.getcwd(), "tests", "files", "1sided.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 4
        card = self.review_controller().card
        assert card.id == "9cff728f"
        assert "question" in card.question()
        filename = os.path.join(os.getcwd(), "tests", "files", "1sided.xml")
        self.xml_importer().do_import(filename)
        assert last_error.startswith("These cards seem to have been imported before")

    def test_card_type_2(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "2sided.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 2
        card_1 = self.review_controller().card
        assert "question" in card_1.question()
        assert "answer" in card_1.answer()
        cards = self.database().cards_from_fact(card_1.fact)
        if cards[0] == card_1:
            card_2 = cards[1]
        else:
            card_2 = cards[0]
        assert "question" in card_2.answer()
        assert "answer" in card_2.question()
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_CARD, )).fetchone()[0] == 2

    def test_card_type_3(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "3sided.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 2
        card_1 = self.review_controller().card
        assert card_1.fact.data == {"f": "f", "p_1": "p", "m_1": "t"}
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_CARD, )).fetchone()[0] == 2

    def test_card_type_3_corrupt(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "3sided_corrupt.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 2
        card_1 = self.review_controller().card
        assert card_1.fact.data == {"f": "f", "m_1": "t"}
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_CARD, )).fetchone()[0] == 2

    def test_card_type_3_missing(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "3sided_missing.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 1
        card_1 = self.review_controller().card
        assert card_1.fact.data == {"f": "t", "b": "f\np"}
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_CARD, )).fetchone()[0] == 1

    def test_media(self):
        global answer
        answer = 0
        os.mkdir(os.path.join(os.getcwd(), "tests", "files", "figs"))
        os.mkdir(os.path.join(os.getcwd(), "tests", "files", "figs", "figs"))
        figures = [\
            os.path.join(os.getcwd(), "tests", "files", "a.png"),
            os.path.join(os.getcwd(), "tests", "files", "figs", "a.png"),
            os.path.join(os.getcwd(), "tests", "files", "figs", "figs", "a.png")]
        for filename in figures:
            open(filename, "w")
        filename = os.path.join(os.getcwd(), "tests", "files", "media.xml")
        self.xml_importer().do_import(filename)
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "a.png"))
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "figs", "a.png"))
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "figs", "a.png"))
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_MEDIA_FILE, )).fetchone()[0] == 3

    def test_media_missing(self):
        global answer
        answer = 0
        os.mkdir(os.path.join(os.getcwd(), "tests", "files", "figs"))
        os.mkdir(os.path.join(os.getcwd(), "tests", "files", "figs", "figs"))
        figures = [\
            os.path.join(os.getcwd(), "tests", "files", "a.png"),
            os.path.join(os.getcwd(), "tests", "files", "figs", "a.png")]
        for filename in figures:
            open(filename, "w")
        filename = os.path.join(os.getcwd(), "tests", "files", "media.xml")
        self.xml_importer().do_import(filename)
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "a.png"))
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "figs", "a.png"))
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_MEDIA_FILE, )).fetchone()[0] == 2

    def test_media_missing_2(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "media.xml")
        self.xml_importer().do_import(filename)
        assert not os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "a.png"))
        assert not os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "figs", "a.png"))
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_MEDIA_FILE, )).fetchone()[0] == 0

    def test_media_slashes(self):
        global answer
        answer = 0
        os.mkdir(os.path.join(os.getcwd(), "tests", "files", "figs"))
        os.mkdir(os.path.join(os.getcwd(), "tests", "files", "figs", "figs"))
        figures = [\
            os.path.join(os.getcwd(), "tests", "files", "a.png"),
            os.path.join(os.getcwd(), "tests", "files", "figs", "a.png"),
            os.path.join(os.getcwd(), "tests", "files", "figs", "figs", "a.png")]
        for filename in figures:
            open(filename, "w")
        filename = os.path.join(os.getcwd(), "tests", "files", "media_slashes.xml")
        self.xml_importer().do_import(filename)
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "a.png"))
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "figs", "a.png"))
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "figs", "a.png"))
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_MEDIA_FILE, )).fetchone()[0] == 3

    def test_sound(self):
        global answer
        answer = 0
        os.mkdir(os.path.join(\
            os.getcwd(), "tests", "files", "soundfiles"))
        soundname = os.path.join(os.path.join(\
            os.getcwd(), "tests", "files", "soundfiles", "a.ogg"))
        open(soundname, "w")
        filename = os.path.join(os.getcwd(), "tests", "files", "sound.xml")
        self.xml_importer().do_import(filename)
        assert os.path.exists(os.path.join(\
            os.path.abspath("dot_test"), "default.db_media", "soundfiles", "a.ogg"))
        assert self.database().con.execute(\
            "select count() from log where event_type=?",
            (EventTypes.ADDED_MEDIA_FILE, )).fetchone()[0] == 1
        self.review_controller().reset()
        card = self.review_controller().card
        assert card.fact["f"] == """<audio src="soundfiles/a.ogg">"""

    def test_map(self):
        filename = os.path.join(os.getcwd(), "tests", "files", "map.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.database().card_count() == 2
        card = self.review_controller().card
        assert card.fact["loc"] == "<b>Drenthe</b>"
        assert card.fact["marked"] == \
          """<img src_missing="maps/Netherlands-Provinces/Drenthe.png">"""
        assert card.fact["blank"] == \
          """<img src_missing="maps/Netherlands-Provinces/Netherlands-Provinces.png">"""

    def test_dups(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "dups.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        assert self.review_controller().card.fact["loc"] == \
               """<b>Freistaat Th\xfcringen (Free State of Thuringia)</b>"""
        assert self.review_controller().card.tag_string() == "Germany: States, MISSING_MEDIA"

    def test_anon_id(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "anon_id.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        fact = self.review_controller().card.fact
        for card in self.database().cards_from_fact(fact):
            assert not card.id.startswith("_")
        assert self.database().card_count() == 2

    def test_anon_id_2(self):
        global answer, last_error
        answer = 0
        last_error = None
        filename = os.path.join(os.getcwd(), "tests", "files", "anon_id.xml")
        self.xml_importer().do_import(filename)
        assert last_error == None
        self.xml_importer().do_import(filename)
        assert last_error == None
        self.review_controller().reset()
        fact = self.review_controller().card.fact
        for card in self.database().cards_from_fact(fact):
            assert not card.id.startswith("_")
        assert self.database().card_count() == 4

    def test_no_id(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "no_id.xml")
        self.xml_importer().do_import(filename)
        self.review_controller().reset()
        fact = self.review_controller().card.fact
        for card in self.database().cards_from_fact(fact):
            assert card.id

    def test_bad_xml(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "bad_xml.xml")
        self.xml_importer().do_import(filename)
        assert last_error.startswith("Unable to parse")

    def test_tags(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "tag.xml")
        self.xml_importer().do_import(filename, extra_tag_names="extra")
        self.review_controller().reset()
        assert len(self.review_controller().card.tags) == 2

    def test_log(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "sound.xml")
        self.xml_importer().do_import(filename)
        ids = [cursor[0] for cursor in self.database().con.execute(\
            "select distinct object_id from log where event_type='6' or event_type='7'")]
        assert ids == ['ef2e21e1']

    def teardown(self):
        global answer
        answer = 0
        filename = os.path.join(os.getcwd(), "tests", "files", "basedir_bz2",
                                    "DIRECTORY_NO_LONGER_USED_BY_MNEMOSYNE2")
        if os.path.exists(filename):
            os.remove(filename)
        filename = os.path.join(os.getcwd(), "tests", "files", "a.png")
        if os.path.exists(filename):
            os.remove(filename)
        filename = os.path.join(os.getcwd(), "tests", "files", "a.ogg")
        if os.path.exists(filename):
            os.remove(filename)
        dirname = os.path.join(os.getcwd(), "tests", "files", "figs")
        if os.path.exists(dirname):
            shutil.rmtree(dirname)
        dirname = os.path.join(os.getcwd(), "tests", "files", "soundfiles")
        if os.path.exists(dirname):
            shutil.rmtree(dirname)
        MnemosyneTest.teardown(self)



#
# SQLite_media.py <Peter.Bienstman@UGent.be>
#

import os
import re
try:
    from hashlib import md5
except ImportError:
    from md5 import md5

if "ANDROID" in os.environ:
    from mnemosyne.android_python.utf8_filenames import *

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.utils import normalise_path
from mnemosyne.libmnemosyne.utils import expand_path, contract_path
from mnemosyne.libmnemosyne.utils import is_filesystem_case_insensitive
from mnemosyne.libmnemosyne.utils import copy_file_to_dir, remove_empty_dirs_in

re_src = re.compile(r"""(src|data)=\"(.+?)\"""", re.DOTALL | re.IGNORECASE)


class SQLiteMedia(object):

    """Code to be injected into the SQLite database class through inheritance,
    so that SQLite.py does not becomes too large.


    """

    def media_dir(self):
        if self.config()["last_database"] == \
            os.path.basename(self.config()["last_database"]):
            return os.path.join(self.config().data_dir,
                os.path.basename(self.config()["last_database"]) + "_media")
        else:
            return self.config()["last_database"] + "_media"

    def create_media_dir_if_needed(self):
        media_dir = self.media_dir()
        if not os.path.exists(media_dir):
            try:
                os.makedirs(media_dir)
            except OSError:
                self.main_widget().show_error(_("Could not create" ) + " " + \
media_dir + ".\n" + _("Check your file permissions and make sure the directory is not open in a file browser."))

    def fact_contains_static_media_files(self, fact):
        # Could be part of fact.py, but is put here to have all media related
        # functions in one place.
        return re_src.search("".join(fact.data.values())) != None

    def _media_hash(self, filename):

        """A hash function that will be used to determine whether or not a
        media file has been modified outside of Mnemosyne.

        'filename' is a relative path inside the media dir.

        In principle, you could have different hash implementations on
        different systems, as the hash is considered something internal and is
        not sent across during sync e.g..

        """

        filename = normalise_path(os.path.join(self.media_dir(), filename))
        if not os.path.exists(filename):
            return "0"
        media_file = open(filename, "rb")
        hasher = md5()
        while True:
            buffer = media_file.read(8096)
            if not buffer:
                break
            hasher.update(buffer)
        return hasher.hexdigest()

        # The following implementation uses the modification date. Less
        # robust, but could be useful on a mobile device.

        #return os.path.getmtime(media_file)

    def check_for_edited_media_files(self):
        # Regular media files.
        new_hashes = {}
        for sql_res in self.con.execute("select filename, _hash from media"):
            filename, hash = normalise_path(sql_res[0]), sql_res[1]
            if not os.path.exists(expand_path(filename, self.media_dir())):
                continue
            new_hash = self._media_hash(filename)
            if hash != new_hash:
                new_hashes[filename] = new_hash
        for filename, new_hash in new_hashes.items():
            self.con.execute("update media set _hash=? where filename=?",
                (new_hash, filename))
            self.log().edited_media_file(filename)

    def dynamically_create_media_files(self):
        # First check which components are actually working. E.g., on a
        # headless server, it's possible that latex is not installed, so
        # we don't need to go through all the effort.
        creators = [f for f in self.component_manager.all("hook",
            "dynamically_create_media_files") if f.is_working() == True]
        if len(creators) == 0:
            return
        for cursor in self.con.execute("select value from data_for_fact"):
            for creator in creators:
                creator.run(cursor[0])

    def active_dynamic_media_files(self):
        # Other media files, e.g. latex.
        filenames = set()
        for hook in self.component_manager.all\
            ("hook", "active_dynamic_media_files"):
            # Prefilter data we need to screen.
            sql_command = """select value from data_for_fact where _fact_id in
                (select _fact_id from cards where active=1) and ("""
            for tag in hook.tags:
                sql_command += """value like '%""" + tag + """%' or """
            sql_command = sql_command[:-3] + ")"
            for cursor in self.con.execute(sql_command):
                filenames = filenames.union(hook.run(cursor[0]))
        return filenames

    def _process_media(self, fact):

        """Copy the media files to the media directory and edit the media
        table. We don't keep track of which facts use which media and delete
        a media file if it's no longer in use. The reason for this is that some
        people use the media directory as their only location to store their
        media files, and also use these files for other purposes.

        Note that not all 'added_media_file' events originated here, they are
        also generated by the latex subsystem, or by checking for files which
        were modified outside of Mnemosyne.

        """

        for match in re_src.finditer("".join(fact.data.values())):
            filename = match.group(2)
            if filename.startswith("http:"):
                continue
            if len(filename) > 200:
                self.main_widget().show_information(\
_("Media filename rather long. This could cause problems using this file on a different OS."))
            if "#" in filename:
                self.main_widget().show_information(\
_("Filename contains '#', which could cause problems on some operating systems."))
            if not os.path.exists(filename) and \
                not os.path.exists(expand_path(filename, self.media_dir())):
                self.main_widget().show_error(\
                    _("Missing media file!") + "\n\n" + filename)
                for fact_key, value in fact.data.items():
                    fact.data[fact_key] = \
                        fact.data[fact_key].replace(match.group(),
                        "src_missing=\"%s\"" % match.group(2))
                continue
            # If needed, copy file to the media dir. Normally this happens when
            # the user clicks 'Add image' e.g., but he could have typed in the
            # full path directly.
            if os.path.isabs(filename):
                filename = copy_file_to_dir(filename, self.media_dir())
            else:  # We always store Unix paths internally.
                filename = filename.replace("\\", "/")
            for fact_key, value in fact.data.items():
                fact.data[fact_key] = value.replace(match.group(2), filename)
                self.con.execute("""update data_for_fact set value=? where
                    _fact_id=? and key=?""",
                    (fact.data[fact_key], fact._id, fact_key))
            if self.con.execute("select 1 from media where filename=? limit 1",
                                (filename, )).fetchone() is None:
                self.con.execute("""insert into media(filename, _hash)
                    values(?,?)""", (filename, self._media_hash(filename)))
                # When we are applying log entries during sync or import, the
                # side effects of e.g. ADDED_FACT events should not generate
                # additional ADDED_MEDIA_FILE events at the remote partner, so
                # we disable the logging of these side effects in that case.
                if not self.syncing and not self.importing:
                    self.log().added_media_file(filename)

    def unused_media_files(self):

        """Returns a set of media files which are in the media directory but
        which are not referenced in the cards.

        """

        case_insensitive = is_filesystem_case_insensitive()
        # Files referenced in the database.
        files_in_db = set()
        for result in self.con.execute(\
            "select value from data_for_fact where value like '%src=%'"):
            for match in re_src.finditer(result[0]):
                filename = match.group(2)
                if case_insensitive:
                    filename = filename.lower()
                files_in_db.add(filename)
        # Files in the media dir.
        files_in_media_dir = set()
        for root, dirnames, filenames in os.walk(self.media_dir()):
            root = contract_path(root, self.media_dir())
            if root.startswith("_"):
                continue
            for filename in filenames:
                # Paths are stored using unix convention.
                if root:
                    filename = root + "/" + filename
                if case_insensitive:
                    filename = filename.lower()
                files_in_media_dir.add(filename)
        return files_in_media_dir - files_in_db

    def delete_unused_media_files(self, unused_files):

        """Delete media files which are no longer in use. 'unused_files'
        should be a subset of 'self.unused_media_files', because here we no
        longer check if these media files are used or not.

        """
        for filename in unused_files:
            os.remove(expand_path(filename, self.media_dir()))
            self.log().deleted_media_file(filename)
        # Purge empty dirs.
        for root, dirnames, filenames in \
            os.walk(self.media_dir(), topdown=False):
            contracted_root = contract_path(root, self.media_dir())
            if not contracted_root or contracted_root.startswith("_"):
                continue
            if len(filenames) == 0 and len(dirnames) == 0:
                os.rmdir(root)
        # Other media files, e.g. latex.
        for f in self.component_manager.all("hook",
            "delete_unused_media_files"):
            f.run()
        remove_empty_dirs_in(self.media_dir())




#
#  smconv_XML.py <Felix.Engel@fcenet.de>, <Peter.Bienstman@UGent.be>
#

import os
import math
import time
from xml.etree import cElementTree

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.file_format import FileFormat
from mnemosyne.libmnemosyne.file_formats.media_preprocessor \
    import MediaPreprocessor

HOUR = 60 * 60 # Seconds in an hour.
DAY = 24 * HOUR # Seconds in a day.


class Smconv_XML(FileFormat, MediaPreprocessor):

    """Import the xml file created by the smconv.pl script to Mnemosyne.
    smconv.pl is available at http://smconvpl.sourceforge.net and reads
    SuperMemo for Palm databases and exports them to XML.

    In order to import the generated XML into mnemosyne, care must be taken
    to ensure the correct charset encoding of the input file. In my case,
    the palm databases are "windows-1252". The xml file generated by smconv.pl
    was set to "us-ascii". This makes the XML parser fail. For me, changing
    the xml header to <?xml version="1.0" encoding="windows-1252"?>  worked
    well. However, your mileage may vary.

    Restrictions:

    - SM for Palm has six fields for each card. Templates can be used to
      format these fields and to control whether they are part of  the
      question or of the answer. However this class assumes that the first
      field is the question and the second field is the answer.

    """

    description = _("Supermemo for Palm through smconv.pl")
    extension = ".xml"
    filename_filter = _("Supermemo for Palm XML files (*.xml *.XML)")
    import_possible = True
    export_possible = False

    def __init__(self, component_manager):
        FileFormat.__init__(self, component_manager)
        MediaPreprocessor.__init__(self, component_manager)

    def do_import(self, filename, extra_tag_names=""):
        FileFormat.do_import(self, filename, extra_tag_names)
        w = self.main_widget()
        try:
            tree = cElementTree.parse(filename)
        except cElementTree.ParseError as e:
            w.show_error(_("Unable to parse file:") + str(e))
            return
        card_type = self.card_type_with_id("1")
        tag_names = [tag_name.strip() for \
            tag_name in extra_tag_names.split(",") if tag_name.strip()]
        for element in tree.find("cards").findall("card"):
            category = element.attrib["category"]
            commit = not (element.attrib["commit"] == "0")
            for field in element.find("card_fields").findall("card_field"):
                if field.attrib["idx"] == "1":
                    question = field.text
                else:
                    answer = field.text
            card_other = element.find("card_other")
            if card_other is None:
                difficulty = 40
                difficulty_prev = 40
            else:
                difficulty = int(card_other.attrib["difficulty"])
                difficulty_prev = int(card_other.attrib["difficulty_prev"])
            # Grades are 0-5. In SM for Palm there are commited and uncommited
            # cards. Uncommited cards go to grade -1.
            # Otherwise try to extrapolate something from difficulty in SM
            # I have implemented guess_grade such, that the distribution of
            # grades looks reasonable for my test database of 4000 entries.
            # By "reasonable" I mean than most of the entries should be
            # at grade 4. I've been learning that database for 4 years, so the
            # cards should have converged by now.
            if commit == False:
                grade = -1
            # Very easy items are scarce in SM and must be easiest grade.
            elif difficulty < 10:
                grade = 5
            # Assign passing grades, based upon whether the difficulty has
            # changed.
            elif difficulty > difficulty_prev:
                grade = 2
            elif difficulty == difficulty_prev:
                grade = 3
            elif difficulty < difficulty_prev:
                grade = 4
            # If the interval becomes shorter, it must have been a failure.
            if card_other is None:
                interval = 0
                interval_prev = 0
            else:
                interval = int(card_other.attrib["interval"]) * DAY
                interval_prev = int(card_other.attrib["interval_prev"]) * DAY
            if interval < interval_prev:
                grade = 0
            # Construct card.
            fact_data = {"f": question, "b": answer}
            self.preprocess_media(fact_data, tag_names)
            card = self.controller().create_new_cards(fact_data, card_type,
                grade=grade, tag_names=tag_names + [category],
                check_for_duplicates=False, save=False)[0]
            if _("MISSING_MEDIA") in tag_names:
                tag_names.remove(_("MISSING_MEDIA"))
            if card_other is not None:
                card.creation_time = int(time.mktime(time.strptime(\
                    card_other.attrib["datecreate"], "%Y-%m-%d")))
                card.modification_time = int(time.mktime(time.strptime(\
                    card_other.attrib["datecommit"], "%Y-%m-%d")))
                card.next_rep = self.scheduler().midnight_UTC(int(time.mktime(\
                    time.strptime(card_other.attrib["datenexttest"],
                    "%Y-%m-%d"))))
                card.last_rep = card.next_rep - interval
                card.lapses = int(card_other.attrib["lapses"])
                # Try to fill acquisiton reps and retention reps.
                # Since SM statistics are only available for commited
                # cards, I take acq_reps = 0 and ret_reps = lapses + recalls.
                card.ret_reps = card.lapses + int(card_other.attrib["recalls"])
                # Try to derive an easines factor EF from [1.3 .. 3.2] from
                # difficulty d from [1% .. 100%].
                # The math below is set to translate
                # difficulty=100% --> easiness = 1.3
                # difficulty=40% --> easiness = 2.5
                # difficulty=1% --> easiness = 3.2
                dp = difficulty * 0.01
                # Small values should be easy, large ones hard.
                if dp > 0.4:
                    card.easiness = 1.28 - 1.32 * math.log(dp)
                else:
                    card.easiness = 4.2 - 1.139 * math.exp(dp)
                self.database().update_card(card)
        self.warned_about_missing_media = False


#
# portuguese.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Portuguese(Language):

    name = _("Portuguese")
    used_for = "pt"
    sublanguages = {"pt-PT": _("Portugese (Portugal)"),
                     "pt-BR": _("Portugese (Brazilian)")}
    feature_description = _("Google translation and text-to-speech.")



#
# binary_format.py <Peter.Bienstman@UGent.be>
#

class BinaryFormat(object):

    """Used when doing the initial sync and downloading the entire database
    as a binary file in order to speed up the sync.

    """

    def __init__(self, database):
        raise NotImplementedError

    def supports(self, program_name, program_version, database_version):
        raise NotImplementedError

    def binary_filename(self, store_pregenerated_data,
            interested_in_old_reps):
        raise NotImplementedError

    def clean_up(self):
        pass



#
# arabic.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Arabic(Language):

    name = _("Arabic")
    used_for = "ar"
    feature_description = _("Google translation and text-to-speech.")



# Mnemosyne configuration file.

# Align question/answers to the left (True/False)
left_align = False

# Keep detailed logs (True/False).
keep_logs = True

# Upload server. Only change when prompted by the developers.
upload_server = "mnemosyne-proj.dyndns.org:80"

# Set to True to prevent you from accidentally revealing the answer
# when clicking the edit button.
only_editable_when_answer_shown = False

# The translation to use, e.g. 'de' for German (including quotes).
# See http://www.mnemosyne-proj.org/help/translations.php for a list
# of available translations.
# If locale is set to None, the system's locale will be used.
locale = None

# The number of daily backups to keep. Set to -1 for no limit.
max_backups = 5

# The moment the new day starts. Defaults to 3 am. Could be useful to
# change if you are a night bird.
day_starts_at = 3



#
# SQLite_sync.py - Peter Bienstman <Peter.Bienstman@UGent.be>
#                  Max Usachev <maxusachev@gmail.com>
#                  Ed Bartosh <bartosh@gmail.com>
#

import os
import re
import time
import sqlite3

from openSM2sync.log_entry import LogEntry
from openSM2sync.log_entry import EventTypes

from mnemosyne.libmnemosyne.tag import Tag
from mnemosyne.libmnemosyne.fact import Fact
from mnemosyne.libmnemosyne.card import Card
from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.libmnemosyne.fact_view import FactView
from mnemosyne.libmnemosyne.utils import MnemosyneError
from mnemosyne.libmnemosyne.utils import normalise_path, expand_path


re_src = re.compile(r"""(src|data)=\"(.+?)\"""", re.DOTALL | re.IGNORECASE)

# Simple named-tuple like class, to avoid the expensive creation a full card
# object (Python 2.5 does not yet have a named tuple).

class Bunch:
    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class SQLiteSync(object):

    """Code to be injected into the SQLite database class through inheritance,
    so that SQLite.py does not becomes too large.

    Note that we only store events like EDITED_CARD, not the data itself that
    was edited. This is filled out during the actual sync with the latest data
    for that card. This needs special care, e.g. if objects are created and
    immediately destroyed, or if information is needed that only becomes
    available later during the sync protocol. It does however save considerable
    space, and since backing up the file before sync can on some platforms be
    the most time consuming step of the sync, we have chosen this option.

    """

    def append_to_sync_partner_info(self, partner_info):
        return partner_info

    def set_sync_partner_info(self, info):
        self.sync_partner_info = info
        # We also use this function to set some variables at the beginning of
        # the sync process.
        self.reapply_default_criterion_needed = False
        # At the beginning of the sync process, we install a mechanism to make
        # make sure that all card types used during sync can actually be
        # instantiated. This is complicated by the fact that in one sync
        # session, a card can be created, followed by a creation of a new card
        # type, followed by a conversion of that card to that new card type.
        # In this case, the card creation event will already list the new card
        # type, which will only be created later during sync.
        self.card_types_to_instantiate_later = set()

    def partners(self):
        return [cursor[0] for cursor in self.con.execute("""select partner
            from partnerships where partner!=?""", ("log.txt", ))]

    def create_if_needed_partnership_with(self, partner):
        sql_res = self.con.execute("""select partner from partnerships
           where partner=?""", (partner, )).fetchone()
        if not sql_res:
            self.con.execute("""insert into partnerships(partner,
               _last_log_id) values(?,?)""", (partner, 0))

    def remove_partnership_with(self, partner):
        self.con.execute("delete from partnerships where partner=?",
            (partner, ))

    def merge_partners(self, remote_partners):

        """Remember the indirect sync partners. Since we don't need their
        _last_log_id, we set it -1.

        """

        local_partners = self.partners()
        for partner in remote_partners:
            if partner not in local_partners and \
               partner != self.sync_partner_info["machine_id"] and \
               partner != self.config().machine_id():
                self.con.execute("""insert into partnerships(partner,
                     _last_log_id) values(?,?)""", (partner, -1))

    def reset_partnerships(self):

        """Reset partnerships, e.g. after restoring from a backup or after
        doing a full sync. This will make sure that a next sync will be a
        full sync if appropriate.

        """

        self.con.execute(\
            "update partnerships set _last_log_id=? where partner!=?",
            (-666, "log.txt"))

    def is_sync_reset_needed(self, partner):
        return self.last_log_index_synced_for(partner) == -666

    def last_log_index_synced_for(self, partner):
        return self.con.execute("""select _last_log_id from partnerships
           where partner=?""", (partner, )).fetchone()[0]

    def update_last_log_index_synced_for(self, partner):
        # We use this function to do some cleaning up at the end of the sync
        # process as well.
        #
        # At the end of the sync, see if we were able to actually instantiate
        # all card types.
        if len(self.card_types_to_instantiate_later) != 0:
            raise RuntimeError(_("Missing plugins for card types."))
        # See if we need to reapply the default criterion.
        if self.reapply_default_criterion_needed:
            criterion = self.current_criterion()
            applier = self.component_manager.current("criterion_applier",
                used_for=criterion.__class__)
            applier.apply_to_database(criterion)
        # Now we can update the last log index.
        self.con.execute(\
            "update partnerships set _last_log_id=? where partner=?",
            (self.current_log_index(), partner))

    def number_of_log_entries_to_sync_for(self, partner,
            interested_in_old_reps=True):
        _id = self.last_log_index_synced_for(partner)
        if interested_in_old_reps:
            return self.con.execute("select count() from log where _id>?",
                (_id, )).fetchone()[0]
        else:
            return self.con.execute("""select count() from log where _id>? and
                event_type!=?""", (_id, EventTypes.REPETITION)).fetchone()[0]

    def number_of_log_entries(self, interested_in_old_reps=True):
        if interested_in_old_reps:
            return self.con.execute("select count() from log").fetchone()[0]
        else:
            return self.con.execute("""select count() from log where
                event_type!=?""", (EventTypes.REPETITION,)).fetchone()[0]

    def log_entries_to_sync_for(self, partner, interested_in_old_reps=True):

        """Note that we return an iterator here to be able to stream
        efficiently.

        """

        _id = self.last_log_index_synced_for(partner)
        if interested_in_old_reps:
            return (self._log_entry(cursor) for cursor in self.con.execute(\
                "select * from log where _id>?", (_id, )))
        else:
            return (self._log_entry(cursor) for cursor in self.con.execute(\
                "select * from log where _id>? and event_type!=?",
                (_id, EventTypes.REPETITION)))

    def all_log_entries(self, interested_in_old_reps=True):
        if interested_in_old_reps:
            return (self._log_entry(cursor) for cursor in self.con.execute(\
                "select * from log"))
        else:
            return (self._log_entry(cursor) for cursor in self.con.execute(\
                "select * from log where event_type!=?",
                (EventTypes.REPETITION, )))

    def media_filenames_to_sync_for(self, partner):

        """Determine which media files need to be sent across during the sync.
        Obviously, this only includes existing media files, not deleted ones.

        """

        _id = self.last_log_index_synced_for(partner)
        filenames = set()
        for filename in [cursor[0] for cursor in self.con.execute(\
            """select object_id from log where _id>? and (event_type=? or
            event_type=?)""", (_id, EventTypes.ADDED_MEDIA_FILE,
            EventTypes.EDITED_MEDIA_FILE))]:
            if os.path.exists(\
                normalise_path(expand_path(filename, self.media_dir()))):
                filenames.add(filename)
        return filenames

    def all_media_filenames(self):

        """Determine all media files, for use in the initial full sync."""

        # We cannot rely on logs in the database here, since part of it
        # may have been archived, so we simply send across the entire
        # media directory.

        filenames = set()
        for root, dirs, files in os.walk(self.media_dir()):
            subdir = root.replace(self.media_dir(), "")
            for name in files:
                filename = os.path.join(subdir, name)
                if filename.startswith("\\") or filename.startswith("/"):
                    filename = filename[1:]
                filenames.add(filename)
        return filenames

    def generate_log_entries_for_settings(self):

        """Needed after binary initial upload/download of the database, to
        ensure that the side effects to config get applied.

        """

        for key in self.config().keys_to_sync:
            self.log().edited_setting(key)

    def active_objects_to_export(self):
        active_objects = {}
        # Active facts (working with python sets turns out to be more
        # efficient than a 'distinct' statement in SQL).
        active_objects["_fact_ids"] = list(\
            set([cursor[0] for cursor in self.con.execute(\
            "select _fact_id from cards where active=1")]))
        # Active cards and their inactive sister cards.
        # (We need to log cards too instead of just facts, otherwise we cannot
        # communicate the card type.)
        active_objects["_card_ids"] = \
            [cursor[0] for cursor in self.con.execute(\
            """select _id from cards where _fact_id in
            (select _fact_id from cards where active=1)""")]
        # Tags belonging to the active cards.
        active_objects["tags"] = self.tags_from_cards_with_internal_ids(\
            active_objects["_card_ids"])
        # User defined card types.
        defined_in_database_ids = set([cursor[0] for cursor in \
            self.con.execute("select id from card_types")])
        # Card types of the active cards (no need to include inactive sister
        # cards here, as they share the same card type).
        active_card_type_ids = set([cursor[0] for cursor in self.con.execute(\
            "select card_type_id from cards where active=1")]).\
            intersection(defined_in_database_ids)
        # Also add parent card types, even if they are not active at the
        # moment.
        parent_card_type_ids = set()
        for id in active_card_type_ids:
            while "::" in id: # Move up one level of the hierarchy.
                id, child_name = id.rsplit("::", 1)
                if id in defined_in_database_ids and \
                    id not in active_card_type_ids:
                    parent_card_type_ids.add(id)
        # Sort to make sure we add the parents first.
        active_objects["card_type_ids"] = \
            sorted(active_card_type_ids.union(parent_card_type_ids))
        # Fact views.
        active_objects["fact_view_ids"] = []
        for card_type_id in active_objects["card_type_ids"]:
            active_objects["fact_view_ids"] += eval(self.con.execute(\
                "select fact_view_ids from card_types where id=?",
                (card_type_id, )).fetchone()[0])
        # Media files for active cards.
        active_objects["media_filenames"] = self.active_dynamic_media_files()
        for result in self.con.execute(\
            """select value from data_for_fact where _fact_id in (select
            _fact_id from cards where active=1) and value like '%src=%'"""):
            for match in re_src.finditer(result[0]):
                active_objects["media_filenames"].add(match.group(2))
        return active_objects

    def set_extra_tags_on_import(self, tags):
        self.extra_tags_on_import = tags

    def _log_entry(self, sql_res):

        """Create log entry object in the format openSM2sync expects."""

        log_entry = LogEntry()
        log_entry["type"] = sql_res[1]
        log_entry["time"] = sql_res[2]
        o_id = sql_res[3]
        if o_id:
            log_entry["o_id"] = o_id
        event_type = log_entry["type"]
        if event_type in (EventTypes.LOADED_DATABASE,
            EventTypes.SAVED_DATABASE):
            log_entry["sch"] = sql_res[6]
            log_entry["n_mem"] = sql_res[7]
            log_entry["act"] = sql_res[8]
        elif event_type in (EventTypes.ADDED_CARD, EventTypes.EDITED_CARD):
            if self.has_card_with_id(log_entry["o_id"]):
                # Note that some of these values (e.g. the repetition count) we
                # could in theory calculate from the previous state and the
                # grade. However, we send the entire state of the card across
                # because it could be that there is no valid previous state
                # because of conflict resolution.
                # Note that we deliberately do not send across 'active', as
                # this is controlled by the remote client.
                card = self.card(log_entry["o_id"], is_id_internal=False)
                if self.sync_partner_info.get("capabilities") == "cards":
                    log_entry["f"] = card.question("sync_to_card_only_client")
                    log_entry["b"] = card.answer("sync_to_card_only_client")
                else:
                    log_entry["card_t"] = card.card_type.id
                    log_entry["fact"] = card.fact.id
                    log_entry["fact_v"] = card.fact_view.id
                log_entry["c_time"] = card.creation_time
                log_entry["m_time"] = card.modification_time
                log_entry["tags"] = ",".join([tag.id for tag in card.tags])
                log_entry["gr"] = card.grade
                log_entry["e"] = card.easiness
                log_entry["l_rp"] = card.last_rep
                log_entry["n_rp"] = card.next_rep
                log_entry["ac_rp"] = card.acq_reps
                log_entry["rt_rp"] = card.ret_reps
                log_entry["lps"] = card.lapses
                log_entry["ac_rp_l"] = card.acq_reps_since_lapse
                log_entry["rt_rp_l"] = card.ret_reps_since_lapse
                log_entry["sch_data"] = card.scheduler_data
                if card.extra_data:
                    log_entry["extra"] = repr(card.extra_data)
            else: # The object has been deleted at a later stage.
                pass
        elif event_type == EventTypes.REPETITION:
            log_entry["gr"] = sql_res[4]
            log_entry["e"] = sql_res[5]
            log_entry["sch_i"] = sql_res[11]
            log_entry["act_i"] = sql_res[12]
            log_entry["th_t"] = sql_res[13]
            log_entry["ac_rp"] = sql_res[6]
            log_entry["rt_rp"] = sql_res[7]
            log_entry["lps"] = sql_res[8]
            log_entry["ac_rp_l"] = sql_res[9]
            log_entry["rt_rp_l"] = sql_res[10]
            log_entry["n_rp"] = sql_res[14]
            log_entry["sch_data"] = sql_res[15]
        elif event_type in (EventTypes.ADDED_TAG, EventTypes.EDITED_TAG):
            if self.has_tag_with_id(log_entry["o_id"]):
                tag = self.tag(log_entry["o_id"], is_id_internal=False)
                if tag is None:
                    return None
                log_entry["name"] = tag.name
                if tag.extra_data:
                    log_entry["extra"] = repr(tag.extra_data)
            else: # The object has been deleted at a later stage.
                pass
        elif event_type in (EventTypes.ADDED_MEDIA_FILE,
            EventTypes.EDITED_MEDIA_FILE, EventTypes.DELETED_MEDIA_FILE):
            log_entry["fname"] = sql_res[3]
            del log_entry["o_id"]
        elif event_type in (EventTypes.ADDED_FACT, EventTypes.EDITED_FACT):
            if self.sync_partner_info.get("capabilities") == "cards":
                # The accompanying ADDED_CARD and EDITED_CARD events suffice.
                return None
            if self.has_fact_with_id(log_entry["o_id"]):
                fact = self.fact(log_entry["o_id"], is_id_internal=False)
                for fact_key, value in fact.data.items():
                    log_entry[fact_key] = value
            else: # The object has been deleted at a later stage.
                pass
        elif event_type in (EventTypes.ADDED_FACT_VIEW,
            EventTypes.EDITED_FACT_VIEW):
            if self.has_fact_view_with_id(log_entry["o_id"]):
                fact_view = self.fact_view(log_entry["o_id"],
                    is_id_internal=False)
                log_entry["name"] = fact_view.name
                log_entry["q_fact_keys"] = repr(fact_view.q_fact_keys)
                log_entry["a_fact_keys"] = repr(fact_view.a_fact_keys)
                log_entry["q_fact_key_decorators"] = \
                    repr(fact_view.q_fact_key_decorators)
                log_entry["a_fact_key_decorators"] = \
                    repr(fact_view.a_fact_key_decorators)
                log_entry["a_on_top_of_q"] = repr(fact_view.a_on_top_of_q)
                log_entry["type_answer"] = repr(fact_view.type_answer)
                if fact_view.extra_data:
                    log_entry["extra"] = repr(fact_view.extra_data)
        elif event_type in (EventTypes.ADDED_CARD_TYPE,
            EventTypes.EDITED_CARD_TYPE):
            if self.has_card_type_with_id(log_entry["o_id"]):
                card_type = self.card_type(log_entry["o_id"],
                    is_id_internal=False)
                log_entry["name"] = card_type.name
                log_entry["fact_keys_and_names"] = \
                    repr(card_type.fact_keys_and_names)
                log_entry["fact_views"] = repr([fact_view.id for fact_view \
                    in card_type.fact_views])
                log_entry["unique_fact_keys"] = \
                    repr(card_type.unique_fact_keys)
                log_entry["required_fact_keys"] = \
                    repr(card_type.required_fact_keys)
                log_entry["keyboard_shortcuts"] = \
                    repr(card_type.keyboard_shortcuts)
                if card_type.extra_data:
                    log_entry["extra"] = repr(card_type.extra_data)
        elif event_type in (EventTypes.ADDED_CRITERION,
            EventTypes.EDITED_CRITERION):
            if self.has_criterion_with_id(log_entry["o_id"]):
                criterion = self.criterion(log_entry["o_id"],
                    is_id_internal=False)
                log_entry["name"] = criterion.name
                log_entry["criterion_type"] = criterion.criterion_type
                log_entry["data"] = criterion.data_to_sync_string()
        elif event_type == EventTypes.EDITED_SETTING:
            if log_entry["o_id"] not in self.config():  # Obsolete entry.
                return None
            log_entry["value"] = repr(self.config()[log_entry["o_id"]])
        return log_entry

    def add_tag_from_log_entry(self, log_entry):
        already_imported = self.has_tag_with_id(log_entry["o_id"])
        if "name" not in log_entry:
            log_entry["name"] = "dummy"  # Added and immediately deleted.
        same_name_in_database = self.con.execute(\
            "select 1 from tags where name=? and id!=? limit 1",
            (log_entry["name"], log_entry["o_id"])).fetchone() is not None
        if same_name_in_database:
            # Merging with the tag which is already in the database is more
            # difficult, as then the tag links in the cards would need to
            # be updated.
            if self.importing:  # Don't interrupt sync with dialog.
                self.main_widget().show_information(\
        _("Tag '%s' already in database, renaming new tag to '%s (1)'" \
                 % (log_entry["name"], log_entry["name"])))
            log_entry["name"] += " (1)"
        if already_imported and self.importing:
            log_entry["type"] = EventTypes.EDITED_TAG
            return self.update_tag(self.tag_from_log_entry(log_entry))
        self.add_tag(self.tag_from_log_entry(log_entry))

    def tag_from_log_entry(self, log_entry):
        # When deleting, the log entry only contains the tag's id, so we pull
        # the object from the database. This is a bit slower than just filling
        # in harmless missing keys, but it is more robust against future
        # side effects of tag deletion.
        if log_entry["type"] == EventTypes.DELETED_TAG:
            # Work around legacy logs which contain duplicate deletion events.
            if self.has_tag_with_id(log_entry["o_id"]):
                return self.tag(log_entry["o_id"], is_id_internal=False)
            else:
                # Leftover from old bug, should not reoccur.
                self.main_widget().show_information(\
            _("Deleting same tag twice during sync. Inform the developpers."))
                log_entry["name"] = "irrelevant"
                return Tag(log_entry["name"], log_entry["o_id"])
        # If we are creating a tag that will be deleted at a later stage
        # during this sync, we are missing some (irrelevant) information
        # needed to properly create a tag object.
        if "name" not in log_entry:
            log_entry["name"] = "irrelevant"
        # Create tag object.
        tag = Tag(log_entry["name"], log_entry["o_id"])
        if "extra" in log_entry:
            tag.extra_data = eval(log_entry["extra"])
        # Make sure to create _id fields as well, otherwise database
        # operations or their side effects could fail.
        if log_entry["type"] != EventTypes.ADDED_TAG:
            tag._id = self.con.execute("select _id from tags where id=?",
                (tag.id, )).fetchone()[0]
        return tag

    def add_fact_from_log_entry(self, log_entry):
        if self.importing:
            already_imported = self.has_fact_with_id(log_entry["o_id"])
            if already_imported:
                log_entry["type"] = EventTypes.EDITED_FACT
                return self.update_fact(\
                    self.fact_from_log_entry(log_entry))
        self.add_fact(self.fact_from_log_entry(log_entry))

    def fact_from_log_entry(self, log_entry):
        # Work around legacy logs which contain duplicate deletion events.
        # Leftover from old bug, should not reoccur.
        if log_entry["type"] != EventTypes.ADDED_FACT and \
           not self.has_fact_with_id(log_entry["o_id"]):
            self.main_widget().show_information(\
        _("Deleting same fact twice during sync. Inform the developpers."))
            fact = Fact({}, log_entry["o_id"])
            fact._id = -1
            return fact
        # Get fact object to be deleted now.
        if log_entry["type"] == EventTypes.DELETED_FACT:
            return self.fact(log_entry["o_id"], is_id_internal=False)
        # Create fact object.
        fact_data = {}
        for key, value in log_entry.items():
            if key not in ["time", "type", "o_id"]:
                fact_data[key] = value
        fact = Fact(fact_data, log_entry["o_id"])
        if log_entry["type"] != EventTypes.ADDED_FACT:
            fact._id = self.con.execute("select _id from facts where id=?",
                (fact.id, )).fetchone()[0]
        return fact

    def add_card_from_log_entry(self, log_entry):
        if self.importing:
            already_imported = self.has_card_with_id(log_entry["o_id"])
            if already_imported:
                orig_card = self.card(log_entry["o_id"], is_id_internal=False)
                card = self.card_from_log_entry(log_entry)
                card.grade = orig_card.grade
                card.easiness = orig_card.easiness
                card.acq_reps = orig_card.acq_reps
                card.ret_reps = orig_card.ret_reps
                card.lapses = orig_card.lapses
                card.acq_reps_since_lapse = orig_card.acq_reps_since_lapse
                card.ret_reps_since_lapse = orig_card.ret_reps_since_lapse
                card.next_rep = orig_card.next_rep
                card.last_rep = orig_card.last_rep
                card.modification_time = int(time.time())
                if "extra" in log_entry:
                    card.extra_data = eval(log_entry["extra"])
                card.tags = orig_card.tags
                for tag_id in log_entry["tags"].split(","):
                    card.tags.add(self.tag(tag_id, is_id_internal=False))
                for tag in self.extra_tags_on_import:
                    card.tags.add(tag)
                card.card_type = self.card_type(\
                    log_entry["card_t"], is_id_internal=False)
                if self.is_user_card_type(card.card_type):
                    card.fact_view = self.fact_view(\
                        log_entry["fact_v"], is_id_internal=False)
                else:
                    for fact_view in card.card_type.fact_views:
                        if fact_view.id == log_entry["fact_v"]:
                            card.fact_view = fact_view
                            break
                return self.update_card(card)
        self.add_card(self.card_from_log_entry(log_entry))

    def card_from_log_entry(self, log_entry):
        # We should not receive cards with question and answer data, only
        # cards based on facts.
        if "f" in log_entry:
            raise AttributeError
        # Get card object to be deleted now.
        if log_entry["type"] == EventTypes.DELETED_CARD:
            try:
                return self.card(log_entry["o_id"], is_id_internal=False)
            except MnemosyneError:  # There is no fact in the database.
                # We have created and deleted this card since the last sync,
                # so we just return an empty shell.
                card_type = self.card_type_with_id("1")
                fact = Fact({"f": "f", "b": "b"}, id="")
                card = Card(card_type, fact, card_type.fact_views[0],
                    creation_time=0)
                card._id = None # Signals special case to 'delete_card'.
                card.id = log_entry["o_id"]
                return card
        # Create an empty shell of card object that will be deleted later
        # during this sync.
        if "tags" not in log_entry:
            card_type = self.card_type_with_id("1")
            fact = Fact({"f": "f", "b": "b"}, id="")
            card = Card(card_type, fact, card_type.fact_views[0],
                creation_time=0)
            card.id = log_entry["o_id"]
            return card
        # Create card object.
        if "card_t" not in log_entry:
            # Client only supports simple cards.
            card_type = self.card_type_with_id("1")
        else:
            if log_entry["card_t"] not in \
                self.component_manager.card_type_with_id:
                # If the card type is not in the database, it's possible that
                # the data for this card type will follow later during the
                # sync. In that case, create a dummy card type here, which
                # will be corrected by a later edit event. Hovewer, we still
                # need to instantiate this card type later, so that we can
                # catch errors, e.g. due to bad plugins.
                try:
                    self.activate_plugins_for_card_type_with_id\
                        (log_entry["card_t"])
                    card_type = self.card_type_with_id\
                        (log_entry["card_t"])
                except:
                    self.card_types_to_instantiate_later.add(\
                        log_entry["card_t"])
                    card_type = self.card_type_with_id("1")
                    log_entry["fact_v"] = card_type.fact_views[0].id
            else:
                card_type = self.card_type_with_id(log_entry["card_t"])
        fact = self.fact(log_entry["fact"], is_id_internal=False)
        # When importing, set the creation time to the current time.
        if self.importing and not self.importing_with_learning_data:
            log_entry["c_time"] = int(time.time())
            log_entry["m_time"] = int(time.time())
        for fact_view in card_type.fact_views:
            if fact_view.id == log_entry["fact_v"]:
                card = Card(card_type, fact, fact_view,
                    creation_time=log_entry["c_time"])
                break
        for tag_id in log_entry["tags"].split(","):
            if self.has_tag_with_id(tag_id):
                card.tags.add(self.tag(tag_id, is_id_internal=False))
            else:
                # The tag has been deleted later later during the log. Don't
                # worry about it now, this will be corrected by a later
                # EDITED_CARD event.
                pass
        if self.importing:
            if len(self.extra_tags_on_import) != 0:
                card.tags.discard(\
                    self.tag("__UNTAGGED__", is_id_internal=False))
            for tag in self.extra_tags_on_import:
                card.tags.add(tag)
        # Construct rest of card. The 'active' property does not need to be
        # handled here, as default criterion will be applied to the card
        # in the database functions.
        card.id = log_entry["o_id"]
        if (log_entry["type"] != EventTypes.ADDED_CARD) or self.importing:
            if self.has_card_with_id(card.id):
                card._id = self.con.execute("select _id from cards where id=?",
                    (card.id, )).fetchone()[0]
            else:
                # Importing a card for the first time, so it is not yet in the
                # database.
                pass
        card.modification_time = log_entry["m_time"]
        card.grade = log_entry["gr"]
        card.easiness = log_entry["e"]
        card.acq_reps = log_entry["ac_rp"]
        card.ret_reps = log_entry["rt_rp"]
        card.lapses = log_entry["lps"]
        card.acq_reps_since_lapse = log_entry["ac_rp_l"]
        card.ret_reps_since_lapse = log_entry["rt_rp_l"]
        card.last_rep = log_entry["l_rp"]
        card.next_rep = log_entry["n_rp"]
        if "sch_data" in log_entry:
            card.scheduler_data = log_entry["sch_data"]
        if "extra" in log_entry:
            card.extra_data = eval(log_entry["extra"])
        return card

    def apply_repetition(self, log_entry):
        sch_data = log_entry.get("sch_data") # Gives None if not present.
        card = Bunch(id=log_entry["o_id"], grade=log_entry["gr"],
            easiness=log_entry["e"], acq_reps=log_entry["ac_rp"],
            ret_reps=log_entry["rt_rp"], lapses=log_entry["lps"],
            acq_reps_since_lapse=log_entry["ac_rp_l"],
            ret_reps_since_lapse=log_entry["rt_rp_l"],
            last_rep=log_entry["time"], next_rep=log_entry["n_rp"],
            scheduler_data=sch_data)
        self.log().repetition(card, log_entry["sch_i"], log_entry["act_i"],
            log_entry["th_t"])
        self.con.execute("""update cards set grade=?, easiness=?, acq_reps=?,
            ret_reps=?, lapses=?, acq_reps_since_lapse=?,
            ret_reps_since_lapse=?, last_rep=?, next_rep=?, scheduler_data=?
            where id=?""", (card.grade, card.easiness, card.acq_reps,
            card.ret_reps, card.lapses, card.acq_reps_since_lapse,
            card.ret_reps_since_lapse, card.last_rep, card.next_rep,
            card.scheduler_data, card.id))

    def add_media_file(self, log_entry):

        """ADDED_MEDIA_FILE events get created in several places:
        database._process_media, database.check_for_edited_media_files,
        latex, ... . In order to make sure that all of these are treated
        in the same way, we generate an ADDED_MEDIA_FILE event here, and
        prevent _process_media from generating this event through
        self.syncing = True.

        """

        filename = log_entry["fname"]
        full_path = normalise_path(expand_path(filename, self.media_dir()))
        if os.path.exists(full_path):
            self.con.execute("""insert or replace into media(filename, _hash)
                values(?,?)""", (filename, self._media_hash(filename)))
        self.log().added_media_file(filename)

    def edit_media_file(self, log_entry):
        filename = log_entry["fname"]
        self.con.execute("update media set _hash=? where filename=?",
            (self._media_hash(filename), filename))
        self.log().edited_media_file(filename)

    def delete_media_file(self, log_entry):
        # Actually, we cannot take the responsibility to delete a media file,
        # since it would e.g. break the corner case to add a media file,
        # delete it and than add it again.
        pass

        #filename = log_entry["fname"]
        #full_path = expand_path(filename, self.media_dir())
        # The file could have been remotely deleted before it got a chance to
        # be synced, so we need to check if the file exists before deleting.
        #if os.path_exists(full_path):
        #    os.remove(full_path)
        #self.log().deleted_media_file(filename)

    def add_fact_view_from_log_entry(self, log_entry):
        if self.importing:
            already_imported = self.con.execute(\
                "select 1 from fact_views where id=? limit 1",
                (log_entry["o_id"], )).fetchone() is not None
            # No need to rename fact views here, as the user only names the
            # card types.
            if already_imported:
                log_entry["type"] = EventTypes.EDITED_FACT_VIEW
                return self.update_fact_view(\
                    self.fact_view_from_log_entry(log_entry))
        try:
            self.add_fact_view(self.fact_view_from_log_entry(log_entry))
        except sqlite3.IntegrityError:
            # Leftover from old bug, should not reoccur.
            self.main_widget().show_information(\
    _("Creating same fact view twice during sync. Inform the developpers."))

    def fact_view_from_log_entry(self, log_entry):
        # Get fact view object to be deleted now.
        if log_entry["type"] == EventTypes.DELETED_FACT_VIEW:
            return self.fact_view(log_entry["o_id"], is_id_internal=False)
        # Create an empty shell of fact view object that will be deleted later
        # during this sync.
        if "name" not in log_entry:
            return FactView("irrelevant", log_entry["o_id"])
        # Create fact view object.
        fact_view = FactView(log_entry["name"], log_entry["o_id"])
        fact_view.q_fact_keys = eval(log_entry["q_fact_keys"])
        fact_view.a_fact_keys = eval(log_entry["a_fact_keys"])
        fact_view.q_fact_key_decorators = eval(log_entry["q_fact_key_decorators"])
        fact_view.a_fact_key_decorators = eval(log_entry["a_fact_key_decorators"])
        fact_view.a_on_top_of_q = bool(eval(log_entry["a_on_top_of_q"]))
        fact_view.type_answer = bool(eval(log_entry["type_answer"]))
        if "extra" in log_entry:
            fact_view.extra_data = eval(log_entry["extra"])
        return fact_view

    def add_card_type_from_log_entry(self, log_entry):
        already_imported = self.con.execute(\
            "select 1 from card_types where id=? limit 1",
            (log_entry["o_id"], )).fetchone() is not None
        if "name" not in log_entry:
            log_entry["name"] = "dummy"  # Added and immediately deleted.
        same_name_in_database = self.con.execute(\
            "select 1 from card_types where name=? and id!=? limit 1",
            (log_entry["name"], log_entry["o_id"] )).fetchone() is not None
        if same_name_in_database:
            # Merging with the card type which is already in the database
            # is more difficult, as then the card type links in the cards
            # would need to be updated.
            if self.importing:  # Don't interrupt sync with dialog.
                self.main_widget().show_information(\
    _("Card type '%s' already in database, renaming new card type to '%s (1)'" \
                % (log_entry["name"], log_entry["name"])))
            log_entry["name"] += " (1)"
        if already_imported and self.importing:
            log_entry["type"] = EventTypes.EDITED_CARD_TYPE
            return self.update_card_type(\
                self.card_type_from_log_entry(log_entry))
        try:
            card_type = self.card_type_from_log_entry(log_entry)
            self.activate_plugins_for_card_type_with_id(card_type.id)
            self.add_card_type(card_type)
        except sqlite3.IntegrityError:
            # Leftover from old bug, should not reoccur.
            self.main_widget().show_information(\
    _("Creating same card type twice during sync. Inform the developpers."))

    def card_type_from_log_entry(self, log_entry):
        # Get card type object to be deleted now.
        if log_entry["type"] == EventTypes.DELETED_CARD_TYPE:
            return self.card_type(log_entry["o_id"], is_id_internal=False)
        # Create an empty shell of card type object that will be deleted later
        # during this sync.
        if "fact_views" not in log_entry:
            card_type = CardType(self.component_manager)
            card_type.id = log_entry["o_id"]
            card_type.fact_views = []
            card_type.fact_keys_and_names = []
            return card_type
        # Create card type object.
        card_type = CardType(self.component_manager)
        card_type.id = log_entry["o_id"]
        card_type.name = log_entry["name"]
        card_type.fact_keys_and_names = eval(log_entry["fact_keys_and_names"])
        card_type.fact_views = []
        for fact_view_id in eval(log_entry["fact_views"]):
            card_type.fact_views.append(self.fact_view(fact_view_id,
                is_id_internal=False))
        card_type.unique_fact_keys = eval(log_entry["unique_fact_keys"])
        card_type.required_fact_keys = eval(log_entry["required_fact_keys"])
        card_type.keyboard_shortcuts = eval(log_entry["keyboard_shortcuts"])
        if "extra" in log_entry:
            card_type.extra_data = eval(log_entry["extra"])
        return card_type

    def criterion_from_log_entry(self, log_entry):
        # Get criterion object to be deleted now.
        if log_entry["type"] == EventTypes.DELETED_CRITERION:
            return self.criterion(log_entry["o_id"], is_id_internal=False)
        # Create an empty shell of criterion object that will be deleted later
        # during this sync.
        if "criterion_type" not in log_entry:
            from mnemosyne.libmnemosyne.criteria.default_criterion \
                 import DefaultCriterion
            return DefaultCriterion(self.component_manager, log_entry["o_id"])
        # Create criterion object.
        for criterion_class in self.component_manager.all("criterion"):
            if criterion_class.criterion_type == log_entry["criterion_type"]:
                criterion = \
                    criterion_class(self.component_manager, log_entry["o_id"])
                criterion.name = log_entry["name"]
                # Try to apply the data to the criterion. If this fails, it
                # means that the data to do this is not yet available at this
                # stage of the sync, and will be corrected by a later
                # EDITED_CRITERION event.
                try:
                    criterion.set_data_from_sync_string(log_entry["data"])
                except TypeError:
                    pass
        if log_entry["type"] != EventTypes.ADDED_CRITERION:
            criterion._id = self.con.execute("""select _id from criteria where
                id=?""", (criterion.id, )).fetchone()[0]
        return criterion

    def apply_log_entry(self, log_entry, importing=False):
        if not importing:
            self.syncing = True
        else:
            self.importing = True
        event_type = log_entry["type"]
        if "time" in log_entry:
            self.log().timestamp = int(log_entry["time"])
        # TMP measure to allow syncing partners which did not yet store
        # machine ids for LOADED_DATABASE and SAVED_DATABASE.
        if not "o_id" in log_entry:
            log_entry["o_id"] = ""
        try:
            if event_type == EventTypes.STARTED_PROGRAM:
                self.log().started_program(log_entry["o_id"])
            elif event_type == EventTypes.STOPPED_PROGRAM:
                self.log().stopped_program()
            elif event_type == EventTypes.STARTED_SCHEDULER:
                self.log().started_scheduler(log_entry["o_id"])
            elif event_type == EventTypes.LOADED_DATABASE:
                self.log().loaded_database(log_entry["o_id"], log_entry["sch"],
                    log_entry["n_mem"], log_entry["act"])
            elif event_type == EventTypes.SAVED_DATABASE:
                self.log().saved_database(log_entry["o_id"], log_entry["sch"],
                    log_entry["n_mem"], log_entry["act"])
            elif event_type == EventTypes.ADDED_TAG:
                self.add_tag_from_log_entry(log_entry)
            elif event_type == EventTypes.EDITED_TAG:
                self.update_tag(self.tag_from_log_entry(log_entry))
            elif event_type == EventTypes.DELETED_TAG:
                self.delete_tag(self.tag_from_log_entry(log_entry))
            elif event_type == EventTypes.ADDED_FACT:
                self.add_fact_from_log_entry(log_entry)
            elif event_type == EventTypes.EDITED_FACT:
                self.update_fact(self.fact_from_log_entry(log_entry))
            elif event_type == EventTypes.DELETED_FACT:
                self.delete_fact(self.fact_from_log_entry(log_entry))
            elif event_type == EventTypes.ADDED_CARD:
                self.add_card_from_log_entry(log_entry)
            elif event_type == EventTypes.EDITED_CARD:
                self.update_card(self.card_from_log_entry(log_entry))
            elif event_type == EventTypes.DELETED_CARD:
                self.delete_card(self.card_from_log_entry(log_entry))
            elif event_type == EventTypes.REPETITION:
                self.apply_repetition(log_entry)
            elif event_type == EventTypes.ADDED_MEDIA_FILE:
                self.add_media_file(log_entry)
            elif event_type == EventTypes.EDITED_MEDIA_FILE:
                self.edit_media_file(log_entry)
            elif event_type == EventTypes.DELETED_MEDIA_FILE:
                self.delete_media_file(log_entry)
            elif event_type == EventTypes.ADDED_FACT_VIEW:
                self.add_fact_view_from_log_entry(log_entry)
            elif event_type == EventTypes.EDITED_FACT_VIEW:
                self.update_fact_view(self.fact_view_from_log_entry(log_entry))
            elif event_type == EventTypes.DELETED_FACT_VIEW:
                self.delete_fact_view(self.fact_view_from_log_entry(log_entry))
            elif event_type == EventTypes.ADDED_CARD_TYPE:
                self.add_card_type_from_log_entry(log_entry)
                self.card_types_to_instantiate_later.discard(log_entry["o_id"])
            elif event_type == EventTypes.EDITED_CARD_TYPE:
                self.update_card_type(self.card_type_from_log_entry(log_entry))
            elif event_type == EventTypes.DELETED_CARD_TYPE:
                self.delete_card_type(self.card_type_from_log_entry(log_entry))
            elif event_type == EventTypes.ADDED_CRITERION:
                self.add_criterion(self.criterion_from_log_entry(log_entry))
            elif event_type == EventTypes.EDITED_CRITERION:
                criterion = self.criterion_from_log_entry(log_entry)
                self.update_criterion(criterion)
                if criterion.id == "__DEFAULT__":
                    self.reapply_default_criterion_needed = True
            elif event_type == EventTypes.DELETED_CRITERION:
                self.delete_criterion(self.criterion_from_log_entry(log_entry))
            elif event_type == EventTypes.EDITED_SETTING:
                key, value = log_entry["o_id"], eval(log_entry["value"])
                if key in self.config().keys_to_sync:
                    self.config()[key] = value
                    for card_type in self.card_types():
                        for render_chain in self.component_manager.\
                            all("render_chain"):
                            render_chain.renderer_for_card_type(card_type).\
                                update(card_type)
                else:
                    self.log_edited_setting(log_entry["time"], key)
        finally:
            self.log().timestamp = None
            self.syncing = False
            self.importing = False



#
# reverse_arabic.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.filter import Filter
from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.filters.non_latin_font_size_increase import \
    NonLatinFontSizeIncrease

class ReverseArabic(Filter):

    def run(self, text, card, fact_key, **render_args):
        for i in range(len(text)):
            if 0x0590 <= ord(text[i]) <= 0x06FF:
                text = text[::-1]
                break
        return text


class ReverseArabicPlugin(Plugin):

    name = "Reverse Arabic"
    description = "Reverse the Arabic in the web server, to compensate for the fact that the Android browser renders Arabic from left to right."
    components = [ReverseArabic]
    supported_API_level = 3

    def activate(self):
        Plugin.activate(self)
        try:
            self.render_chain("web_server").\
                register_filter(ReverseArabic, in_front=False)
            self.render_chain("web_server").\
                unregister_filter(NonLatinFontSizeIncrease)
        except KeyError:  # The web server chain is not active.
            pass

    def deactivate(self):
        Plugin.deactivate(self)
        try:
            self.render_chain("web_server").\
                unregister_filter(ReverseArabic)
            self.render_chain("web_server").\
                register_filter(NonLatinFontSizeIncrease, in_front=False)
        except KeyError:  # The web server chain is not active.
            pass

# Register plugin.

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(ReverseArabicPlugin)



#
# cloze.py <Peter.Bienstman@UGent.be>
#

import re

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.card import Card
from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.libmnemosyne.fact_view import FactView

cloze_re = re.compile(r"\[(.+?)\]", re.DOTALL)


class Cloze(CardType):

    """CardType to do cloze deletion on a string, e.g. "The political parties in
    the US are the [democrats] and the [republicans]." would give the following
    cards:

    Q:The political parties in the US are the [...] and the republicans.
    A:democrats

    Q:The political parties in the US are the democrats and the [...].
    A:republicans

    Illustration of a CardType which does not use the traditional FactView
    mechanism.

    This is implemented by creating cards which contain extra_data entries
    "cloze" and "index", containing e.g. "democrats" and 0. Storing both the
    cloze and its index allows us to have enough data to support all possible
    editing operations.

    """

    id = "5"
    name = _("Cloze deletion")

    fact_keys_and_names = [("text", _("Text"))]
    unique_fact_keys = ["text"]
    required_fact_keys = ["text"]

    v = FactView(_("Cloze"), "5.1")
    v.q_fact_keys = ["f"]  # Generated on the fly.
    v.a_fact_keys = ["b"]  # Generated on the fly.
    fact_views = [v]

    def fact_key_format_proxies(self):
        return {"text": "text", "f": "text", "b": "text"}

    def is_fact_data_valid(self, fact_data):
        text = fact_data["text"]
        for f in self.component_manager.all("hook", "preprocess_cloze"):
            text = f.run(text)
        return bool(cloze_re.search(text))

    def q_a_from_cloze(self, text, index):

        """Auxiliary function used by other card types to return question
        and answer for the cloze with a given index in a text which can have
        the following form:

        La [casa:house] es [grande:big]

        Use 'index=-1' to get the cloze text without brackets and without
        hints.

        """

        for f in self.component_manager.all("hook", "preprocess_cloze"):
            text = f.run(text)
        cursor = 0
        current_index = 0
        question = text
        answer = None
        while True:
            cursor = text.find("[", cursor)
            if cursor == -1:
                break
            cloze = text[cursor + 1:text.find("]", cursor)]
            if ":" in cloze:
                cloze_without_hint, hint = cloze.split(":", 1)
            else:
                cloze_without_hint, hint = cloze, "..."
            if current_index == index:
                question = question.replace(\
                    "[" + cloze + "]", "[" + hint + "]", 1)
                answer = cloze_without_hint
            else:
                question = question.replace(\
                    "[" + cloze + "]", cloze_without_hint, 1)
            cursor += 1
            current_index += 1
        for f in self.component_manager.all("hook", "postprocess_q_a_cloze"):
            question, answer = f.run(question, answer)
        return question, answer

    def fact_data(self, card):
        question, answer = self.q_a_from_cloze\
            (card.fact["text"], card.extra_data["index"])
        return {"f": question, "b": answer}

    def create_sister_cards(self, fact):
        cards = []
        text = fact["text"]
        for f in self.component_manager.all("hook", "preprocess_cloze"):
            text = f.run(text)
        for match in cloze_re.finditer(text):
            card = Card(self, fact, self.fact_views[0])
            card.extra_data["cloze"] = match.group(1)
            card.extra_data["index"] = len(cards)
            cards.append(card)
        return cards

    def _edit_clozes(self, fact, new_fact_data,
                    cloze_fact_key, cloze_fact_view):

        """Auxiliary function used by other card types to when editing clozes.
        Should take into account that not all fact views are cloze-based.

        """

        new_cards, edited_cards, deleted_cards = [], [], []
        old_clozes = cloze_re.findall(fact[cloze_fact_key])
        new_clozes = cloze_re.findall(new_fact_data[cloze_fact_key])
        # If the number of clozes is equal, just edit the existing cards.
        if len(old_clozes) == len(new_clozes):
            for card in self.database().cards_from_fact(fact):
                if "cloze" in card.extra_data:
                    index = card.extra_data["index"]
                    card.extra_data["cloze"] = new_clozes[index]
                edited_cards.append(card)
        # If not, things are a little more complicated.
        else:
            new_clozes_processed = set()
            for card in self.database().cards_from_fact(fact):
                if "cloze" in card.extra_data:
                    old_cloze = card.extra_data["cloze"]
                    index = card.extra_data["index"]
                    if old_cloze in new_clozes:
                        new_index = new_clozes.index(old_cloze)
                        card.extra_data["cloze"] = new_clozes[new_index]
                        card.extra_data["index"] = new_index
                        new_clozes_processed.add(new_clozes[new_index])
                        edited_cards.append(card)
                    else:
                        deleted_cards.append(card)
                else:
                    edited_cards.append(card)
            for new_cloze in set(new_clozes).difference(new_clozes_processed):
                new_index = new_clozes.index(new_cloze)
                card = Card(self, fact, cloze_fact_view)
                card.extra_data["cloze"] = new_cloze
                card.extra_data["index"] = new_index
                new_cards.append(card)
        return new_cards, edited_cards, deleted_cards

    def edit_fact(self, fact, new_fact_data):
        return self._edit_clozes(fact, new_fact_data,
            "text", self.fact_views[0])


class ClozePlugin(Plugin):

    name = _("Cloze deletion")
    description = _("""A card type blanking out certain fragments in a text.\n
E.g., the text \"The capital of [France] is [Paris]\", will give cards with questions \"The capital of France is [...].\" and \"The capital of [...] is Paris\".\n
Editing the text will automatically update all sister cards.\n\nYou can also specify hints, e.g. [cloze:hint] will show
 [hint] in the question as opposed to [...].""")
    components = [Cloze]
    supported_API_level = 3



#
# utils.py - Peter Bienstman <Peter.Bienstman@UGent.be>

import os
import sys
import math
import random
import locale
import traceback

class SyncError(Exception):
    pass


class SeriousSyncError(Exception):

    """Requires backup from database afterwards."""
    pass


def traceback_string():

    """Like traceback.print_exc(), but returns a string."""

    type, value, tb = sys.exc_info()
    body = "\nTraceback (innermost last):\n"
    list = traceback.format_tb(tb, limit=None) + \
           traceback.format_exception_only(type, value)
    body = body + "%-20s %s" % ("".join(list[:-1]), list[-1])
    del tb  # Prevent circular references.
    return body


def rand_uuid():

    """Importing Python's uuid module brings a huge overhead, so we use
    our own variant: a length 22 random string from a 62 letter alphabet,
    which in terms of randomness is about the same as the traditional hex
    string with length 32, but uses less space.

    """

    chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXZY0123456789"
    rand = random.random
    uuid = ""
    for c in range(22):
        uuid += chars[int(rand() * 62.0 - 1)]
    return uuid


def normalise_path(path):

    """Make sure the correct types of slashes are used.
    'pathlib' itself turns out to be not sufficient for that.

    """

    if os.name == "posix":
        return path.replace("\\", "/")
    else:
        return path.replace("/", "\\")


#
# scheduler.py <Peter.Bienstman@UGent.be>
#

import calendar
import datetime
import time

from mnemosyne.libmnemosyne.component import Component
from mnemosyne.libmnemosyne.gui_translator import _


HOUR = 60 * 60  # Seconds in an hour.
DAY = 24 * HOUR  # Seconds in a day.


class Scheduler(Component):

    name = ""
    component_type = "scheduler"

    def reset(self):

        """Called when starting the scheduler for the first time."""

        raise NotImplementedError

    def set_initial_grade(self, cards, grade):

        """Sets the initial grades for a set of sister cards, making sure
        their next repetitions do no fall on the same day.

        Called when cards are given their initial grade outside of the
        review process, e.g. when the user gives an initial grade when
        adding a new card in the GUI. Therefore, 'unseen' is still left to
        True, as this card has not yet been seen in the interactive review
        process.

        Cards which don't have initial grade information available (e.g. for
        cards created during import or conversion from different card type),
        get their initial grade when they are encountered in the interactive
        review process for the first time.

        In both cases, this initial grading is seen as the first repetition.

        In this way, both types of cards are treated in the same way. (There
        is an ineffectual asymmetry left in the log messages they generate,
        but all the relevant information can still be parsed from them.)

        """

        raise NotImplementedError

    def avoid_sister_cards(self, card):

        """Change card.next_rep to make sure that the card is not scheduled
        on the same day as a sister card.

        Factored out here to allow this to be used by e.g. MnemoGogo.

        """

        raise NotImplementedError

    def rebuild_queue(self, learn_ahead=False):

        """Called by the rest of the library when an existing queue risks
        becoming invalid, e.g. when cards have been deleted in the GUI.
        'next_card' also makes use of this in certain implementations.

        """

        raise NotImplementedError

    def is_in_queue(self, card):

        """To check whether the queue needs to be rebuilt, e.g. if it contains
        a card that was deleted in the GUI.

        """

        raise NotImplementedError

    def remove_from_queue_if_present(self, card):
        raise NotImplementedError

    def next_card(self, learn_ahead=False):
        raise NotImplementedError

    def is_prefetch_allowed(self):

        """Can we display a new card before having processed the grading of
        the previous one?

        """

        raise NotImplementedError

    def grade_answer(self, card, new_grade, dry_run=False):
        raise NotImplementedError

    def scheduled_count(self):
        raise NotImplementedError

    def non_memorised_count(self):
        raise NotImplementedError

    def active_count(self):
        raise NotImplementedError

    def card_count_scheduled_n_days_from_now(self, n):

        """Yesterday: n=-1, today: n=0, tomorrow: n=1, ... .

        Is not implemented in the database, because this could need internal
        scheduler information.
        """

        raise NotImplementedError

    def next_rep_to_interval_string(self, next_rep, now=None):

        """Converts next_rep to a string like 'tomorrow', 'in 2 weeks', ...

        """

        raise NotImplementedError

    def last_rep_to_interval_string(self, last_rep, now=None):

        """Converts last_rep to a string like 'yesterday', '2 weeks ago', ...

        """

        raise NotImplementedError

    def midnight_UTC(self, timestamp):

        """Round a timestamp to a value with resolution of a day, storing it
        in a timezone independent way, as a POSIX timestamp corresponding to
        midnight UTC on that date.

        E.g. if the scheduler sets 'next_rep' to 2012/1/1 12:14 local time,
        this function will return the timestamp corresponding to
        2012/1/1 00;00 UTC.

        Important: the timestamp needs to have the meaning of local time,
        not e.g. UTC, so calling midnight_UTC twice will give unexpected
        results.

        """

        # Create a time tuple containing the local date only, i.e. throwing
        # away hours, minutes, etc.
        # Android/Crystax 10.3.2 actually has a 2038 overflow problem...
        try:
            date_only = datetime.date.fromtimestamp(timestamp).timetuple()
        except OverflowError:
            date_only = datetime.date.fromtimestamp(2**31-2).timetuple()
        # Now we reinterpret this same time tuple as being UTC and convert it
        # to a POSIX timestamp. (Note that timetuples are 'naive', i.e. they
        # themselves do not contain timezone information.)
        return int(calendar.timegm(date_only))

    def adjusted_now(self, now=None):

        """Timezone information and 'day_starts_at' will only become relevant
        when the queue is built, not at schedule time, to allow for
        moving to a different timezone after a card has been scheduled.
        Cards are due when 'adjusted_now >= next_rep', and this function
        makes sure that happens at h:00 local time (with h being
        'day_starts_at').

        """
        
        if now == None:
            now = time.time()
        # The larger 'day_starts_at', the later the card should become due,
        # i.e. larger than 'next_card', so the more 'now' should be decreased.
        now -= self.config()["day_starts_at"] * HOUR
        # 'altzone' or 'timezone' contains the offset in seconds west of UTC.
        # This number is positive for the US, where a card should become
        # due later than in Europe, so 'now' should be decreased by this
        # offset.
        # As for when to use 'altzone' instead of 'timezone' if daylight
        # savings time is active, this is a matter of big confusion
        # among the Python developers themselves:
        # http://bugs.python.org/issue7229
        if time.localtime(now).tm_isdst and time.daylight:
            now -= time.altzone
        else:
            now -= time.timezone 
        return int(now)

    def next_rep_to_interval_string(self, next_rep, now=None):

        """Converts next_rep to a string like 'tomorrow', 'in 2 weeks', ...

        """

        if now is None:
            now = self.adjusted_now()
        interval_days = (next_rep - now) / DAY
        if interval_days >= 365:
            interval_years = interval_days/365.
            return _("in") + " " + "%.1f" % interval_years + " " + \
                   _("years")
        elif interval_days >= 62:
            interval_months = int(interval_days/31)
            return _("in") + " " + str(interval_months) + " " + \
                   _("months")
        elif interval_days >= 31:
            return _("in 1 month")
        elif interval_days >= 1:
            return _("in") + " " + str(int(interval_days) + 1) + " " + \
                   _("days")
        elif interval_days >= 0:
            return _("tomorrow")
        elif interval_days >= -1:
            return _("today")
        elif interval_days >= -2:
            return _("1 day overdue")
        elif interval_days >= -31:
            return str(int(-interval_days)) + " " + _("days overdue")
        elif interval_days >= -62:
            return _("1 month overdue")
        elif interval_days >= -365:
            interval_months = int(-interval_days/31)
            return str(interval_months) + " " + _("months overdue")
        else:
            interval_years = -interval_days/365.
            return "%.1f " % interval_years +  _("years overdue")

    def last_rep_to_interval_string(self, last_rep, now=None):

        """Converts next_rep to a string like 'yesterday', '2 weeks ago', ...

        """

        if now is None:
            now = time.time()
        # To perform the calculation, we need to 'snap' the two timestamps
        # to midnight UTC before calculating the interval.
        now = self.midnight_UTC(\
            now - self.config()["day_starts_at"] * HOUR)
        last_rep = self.midnight_UTC(\
            last_rep - self.config()["day_starts_at"] * HOUR)
        interval_days = (last_rep - now) / DAY
        if interval_days > -1:
            return _("today")
        elif interval_days > -2:
            return _("yesterday")
        elif interval_days > -31:
            return str(int(-interval_days)) + " " + _("days ago")
        elif interval_days > -62:
            return _("1 month ago")
        elif interval_days > -365:
            interval_months = int(-interval_days/31.)
            return str(interval_months) + " " + _("months ago")
        else:
            interval_years = -interval_days/365.
            return "%.1f " % interval_years +  _("years ago")



#
# both_ways.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.card import Card
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.libmnemosyne.fact_view import FactView
from mnemosyne.libmnemosyne.card_type_converter import CardTypeConverter


class BothWays(CardType):

    id = "2"
    name = _("Front-to-back and back-to-front")

    # List and name the keys.
    fact_keys_and_names = [("f", _("Front")),
                           ("b", _("Back"))]

    # Front-to-back.
    v1 = FactView(_("Front-to-back"), "2.1")
    v1.q_fact_keys = ["f"]
    v1.a_fact_keys = ["b"]

    # Back-to-front.
    v2 = FactView(_("Back-to-front"), "2.2",)
    v2.q_fact_keys = ["b"]
    v2.a_fact_keys = ["f"]

    fact_views = [v1, v2]
    required_fact_keys = ["f", "b"]
    unique_fact_keys = ["f"]


from mnemosyne.libmnemosyne.card_types.front_to_back import FrontToBack

class FrontToBackToBothWays(CardTypeConverter):

    used_for = CardTypeConverter.card_type_converter_key\
        (FrontToBack, BothWays)

    def convert(self, cards, old_card_type, new_card_type, correspondence):
        # Update front-to-back view to corresponding view in new type.
        cards[0].fact_view = new_card_type.fact_views[0]
        # Create back-to-front view.
        new_card = Card(new_card_type, cards[0].fact,
            new_card_type.fact_views[1])
        new_cards, edited_cards, deleted_cards = [new_card], [cards[0]], []
        return new_cards, edited_cards, deleted_cards


class BothWaysToFrontToBack(CardTypeConverter):

    used_for = CardTypeConverter.card_type_converter_key\
        (BothWays, FrontToBack)

    def convert(self, cards, old_card_type, new_card_type, correspondence):
        new_cards, edited_cards, deleted_cards = [], [], []
        for card in cards:
            # Update front-to-back view to corresponding view in new type.
            if card.fact_view == old_card_type.fact_views[0]:
                card.fact_view = new_card_type.fact_views[0]
                edited_cards = [card]
            # Delete back-to-front view.
            elif card.fact_view == old_card_type.fact_views[1]:
                deleted_cards = [card]
            else:
                raise ArgumentError("Invalid fact view.")
        return new_cards, edited_cards, deleted_cards



#
# criterion.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.utils import rand_uuid
from mnemosyne.libmnemosyne.component import Component


class Criterion(Component):

    """Used to select a subset of cards, e.g. which cards are currently
    active, i.e. included in the review process.

    The available criteria are stored as classes in the component_manager,
    the actual instances together with their data are stored in the database.

    """

    component_type = "criterion"
    criterion_type = ""
    instantiate = Component.LATER

    def __init__(self, component_manager, id=None):
        Component.__init__(self, component_manager)
        self.name = ""
        if id is None:
            id = rand_uuid()
        self.id = id
        self._id = None

    def __eq__(self, other):
        raise NotImplementedError

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_empty(self):

        """Used to prevent people from creating a criterion which can never
        contain any cards (e.g. disabling all card types).

        """

        return False

    def apply_to_card(self, card):

        """Set the card active or not depending on the criterion. Does not
        write to the database. Called e.g. after creating, updating or
        reviewing cards, to see whether these cards should start out their
        life as active or not.

        The tag and card type creation and deletion functions are callbacks
        called by the rest of libmnemosyne when these objects get created or
        destroyed, such that Criteria can update their status if needed.

        """

        raise NotImplementedError

    def active_tag_added(self, tag):
        pass

    def deactivated_tag_added(self, tag):
        pass

    def is_tag_active(self, tag):
        pass
    
    def tag_deleted(self, tag):
        pass

    def active_card_type_added(self, card_type):
        pass

    def deactivated_card_type_added(self, card_type):
        pass

    def card_type_deleted(self, card_type):
        pass

    def data_to_string(self):

        """Convert variables to a string for storage in the database. We don't
        use pickle here as that would make it difficult for non-Python programs
        to read the database.

        """

        raise NotImplementedError

    def set_data_from_string(self, data_string):
        raise NotImplementedError

    def data_to_sync_string(self):

        """Convert variables to a string for sending across during syncing.
        Could be different from 'data_to_string', as it should use ids instead
        of _ids."""

        raise NotImplementedError

    def set_data_from_sync_string(self, data_string):
        raise NotImplementedError


class CriterionApplier(Component):

    """Can be registered 'used_for' a certain Criterion to apply it in bulk to
    all the cards in the database. Is much faster than fetching each card from
    the database, calling Criterion.apply_to_card, and storing it back in the
    database.

    This code is not part of Criterion, because it is dependent on the database
    backend.

    """

    component_type = "criterion_applier"

    def apply_to_database(self, criterion):
        raise NotImplementedError




#
# export_to_mp3.py <Peter.Bienstman>
#

# Script to export the audio of today's revisions to an mp3 file for hands-off
# review. Only works for mp3 files.

# Adapt it your own need. It uses linux external tools, so it needs to be
# modified to run under Windows.

import re
import os
import time
import shutil
import subprocess

HOUR = 60 * 60 # Seconds in an hour.
DAY = 24 * HOUR # Seconds in a day.

silence = os.path.abspath("silence.mp3")
SILENCE_FACTOR = 1.2 # Silence will be 120% of audio length.

from mnemosyne.script import Mnemosyne

# 'data_dir = None' will use the default sysem location, edit as appropriate.
data_dir = None
mnemosyne = Mnemosyne(data_dir)

# Use mplayer to determine the lenght of an mp3 file.
re_length = re.compile(r"""ID_LENGTH=([0-9]*\.[0-9])""",
    re.DOTALL | re.IGNORECASE)
def determine_length(filename):
    out = subprocess.check_output(["mplayer", "-really-quiet","-vo", "null",
        "-ao", "null", "-frames", "0", "-identify", filename])
    return float(re_length.search(out).group(1))

# Collect all the facts which are either still due or reviewed today.
facts = set()
for cursor in mnemosyne.database().con.execute("""select _id from cards where
    active=1 and grade!=-1 and (?>=next_rep or ?<=last_rep)""",
    (mnemosyne.scheduler().adjusted_now(), time.time() - DAY)):
    card = mnemosyne.database().card(cursor[0], is_id_internal=True)
    facts.add(card.fact)

# For each fact, collect all the mp3 files and add silence near the end.
re_sound = re.compile(r"""audio src=['\"](.+?)['\"]""",
    re.DOTALL | re.IGNORECASE)
subprocess.call(["mp3wrap", "mnemosyne.mp3", silence, silence])
for fact in facts:
    length = 0
    filenames = []
    for match in re_sound.finditer("".join(list(fact.data.values()))):
        filename = os.path.join(\
            mnemosyne.database().media_dir(), match.group(1))
        print((filename, determine_length(filename)))
        length += determine_length(filename)
        filenames.append(filename)
    for i in range(int(length * SILENCE_FACTOR)):
        filenames.append(silence)
    subprocess.call(["mp3wrap", "-a", "mnemosyne_MP3WRAP.mp3"] + filenames)
shutil.move("mnemosyne_MP3WRAP.mp3", "mnemosyne.mp3")



#
# html_css.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.renderer import Renderer

# Css table wizardry based on info from
# http://apptools.com/examples/tableheight.php


class HtmlCss(Renderer):

    """Renders the question or the answer as a full webpage using tables.
    Tested on webkit-based browsers.

    We split out the components of the html page in different functions,
    to allow easier reuse by other renderers.

    """

    used_for = None  # All card types.
    table_height = "100%"

    def __init__(self, component_manager):
        Renderer.__init__(self, component_manager)
        # We cache the css creation to save some time, especially on mobile
        # devices.
        self._css = {} # {card_type.id: render_args: css}

    def body_css(self, **render_args):
        css = "html, body { margin: 0px; height: 100%;  width: 100%;}\n"
        css += "hr { background-color: black; height: 1px; border: 0; }\n"
        return css

    def card_type_css(self, card_type, **render_args):
        # Set aligment of the table (but not the contents within the table).
        # Use a separate id such that user created tables are not affected.
        css = "table.mnem { height: " + self.table_height + "; width: 100%; "
        css += "border: 1px solid #8F8F8F; "
        alignment = self.config().card_type_property(\
            "alignment", card_type, default="center")
        if alignment == "left":
            css += "margin-left: 0; margin-right: auto; "
        elif alignment == "right":
            css += "margin-left: auto; margin-right: 0; "
        else:
            css += "margin-left: auto; margin-right: auto; "
        # Background colours.
        colour = self.config().card_type_property(\
            "background_colour", card_type)
        if colour:
            colour_string = ("%X" % colour)[2:] # Strip alpha.
            css += "background-color: #%s; " % colour_string
        css += "}\n"
        # Key tags.
        for true_fact_key, proxy_fact_key in \
            card_type.fact_key_format_proxies().items():
            css += "div.%s { " % true_fact_key
            # Set alignment within table cell.
            alignment = self.config().card_type_property(\
                "alignment", card_type, proxy_fact_key, default="center")
            css += "text-align: %s; " % alignment
            # Font colours.
            colour = self.config().card_type_property(\
                "font_colour", card_type, proxy_fact_key)
            if colour:
                colour_string = ("%X" % colour)[2:] # Strip alpha.
                css += "color: #%s; " % colour_string
            # Font.
            font_string = self.config().card_type_property(\
                "font", card_type, proxy_fact_key)
            if font_string:
                if font_string.count(",") == 10:
                    family,size,x,x,w,i,u,s,x,x,x = font_string.split(",")
                else:
                    family,size,x,x,w,i,u,s,x,x = font_string.split(",")
                css += "font-family: \"%s\"; " % family
                css += "font-size: %spt; " % size
                if w == "25":
                    css += "font-weight: light; "
                if w == "75":
                    css += "font-weight: bold; "
                if i == "1":
                    css += "font-style: italic; "
                if i == "2":
                    css += "font-style: oblique; "
                if u == "1":
                    css += "text-decoration: underline; "
                if s == "1":
                    css += "text-decoration: line-through; "
            css += "}\n"
        return css

    def update(self, card_type, **render_args):
        if card_type.id not in self._css:
            self._css[card_type.id] = {}
        self._css[card_type.id][repr(sorted(render_args.items()))] =\
            self.body_css(**render_args) + \
            self.card_type_css(card_type, **render_args)

    def css(self, card_type, **render_args):
        render_args_hash = repr(sorted(render_args.items()))
        if not card_type.id in self._css or \
           not render_args_hash in self._css[card_type.id]:
            self.update(card_type, **render_args)
        return self._css[card_type.id][render_args_hash]

    def body(self, fact_data, fact_keys, card_type, **render_args):
        html = ""
        for fact_key in fact_keys:
            if fact_key in fact_data and fact_data[fact_key]:
                line = ""
                if render_args.get("align_top", False):
                    line +="<br>"
                line += "<div id=\"%s\" class=\"%s\">%s</div>" % \
                    (fact_key, fact_key, fact_data[fact_key])
                # Honour paragraph style also in user-created tables.
                line = line.replace("<td>",
                    "<td><div id=\"%s\" class=\"%s\">" % (fact_key, fact_key))
                line = line.replace("<TD>",
                    "<TD><div id=\"%s\" class=\"%s\">" % (fact_key, fact_key))
                line = line.replace("</td>", "</div></td>")
                line = line.replace("</TD>", "</div></TD>")
                html += line
        return html

    def render(self, fact_data, fact_keys, card_type, **render_args):
        css = self.css(card_type)
        valign = "valign=\"top\"" if render_args.get("align_top", False) else ""
        body = self.body(fact_data, fact_keys, card_type, **render_args)
        return """
        <!DOCTYPE html>
        <html>
        <head>
        <meta charset="utf-8">
        <style type="text/css">
        %s
        </style>
        <script type="text/javascript">
          function scroll_to_answer () {
            window.location.hash = "#answer"
          };
        </script>
        </head>
        <body onload="scroll_to_answer()">
          <table id="mnem1" class="mnem">
            <tr %s>
              <td>%s</td>
            </tr>
          </table>
        </body>
        </html>""" % (css, valign, body)

